
#WINDOW_Grep_Width  = 400
#WINDOW_Grep_Height = 170

GrepOutputX = 100
GrepOutputY = 100
GrepOutputWidth = 300
GrepOutputHeight = 300


Procedure CheckGrepAbort()

  Repeat
    Event = WindowEvent()
    
    If EventWindowID() = #WINDOW_Grep
      If EventGadgetID() = #GADGET_Grep_Stop
        Result = 1
      EndIf
    EndIf
    
  Until Event = 0
  
  ProcedureReturn Result
    
EndProcedure


Procedure DisableGrepWindow(State)
  
  DisableGadget(#GADGET_Grep_FindWord, State)

  DisableGadget(#GADGET_Grep_Directory, State)
  DisableGadget(#GADGET_Grep_SelectDirectory, State)
  
  DisableGadget(#GADGET_Grep_Pattern, State)
      
  DisableGadget(#GADGET_Grep_MatchCase, State)
  DisableGadget(#GADGET_Grep_Recurse, State)

  DisableGadget(#GADGET_Grep_Cancel, State)
  DisableGadget(#GADGET_Grep_Stop, 1-State)
  DisableGadget(#GADGET_Grep_Find, State)
    
EndProcedure


; Check if the extension is found in the extension list under the form:
; *.pb, *.pbi, *.pbh
;
Procedure CheckExtension(PatternList$, Extension$)

  If Trim(PatternList$) = ""
    ProcedureReturn #True
  EndIf

  Index = 1
  Repeat
    Pattern$ = StringField(PatternList$, Index, ",")
    Pattern$ = Trim(Pattern$)
    
    If Pattern$      
    
      If Pattern$ = "*" Or Pattern$ = "*.*"
        ProcedureReturn #True
      EndIf
        
      Pattern$ = RemoveString(Pattern$, "*")
      Pattern$ = RemoveString(Pattern$, ".")

      If Trim(Pattern$) = Extension$
        ProcedureReturn #True
      EndIf
    EndIf
        
    Index+1
  Until Pattern$ = ""

EndProcedure


Procedure SearchStringInFile(FileID, Filename$, String$, InitialPathLength, CaseSensitive)

  If ReadFile(FileID, Filename$)
  
    FileLength   = Lof()
    StringLength = Len(String$)
    
    If FileLength And StringLength
  
      *BufferStart = AllocateMemory(FileLength)
      If *BufferStart
      
        ReadData(*BufferStart, FileLength)
        
        Line = 1
        *Buffer.Byte = *BufferStart
        *BufferEnd   = *Buffer+FileLength
        *LineStart   = *BufferStart
        
        While *Buffer < *BufferEnd
        
          If CompareMemoryString(*Buffer, String$, 1-CaseSensitive, StringLength) = 0
            LineLength = 0
            
            ; Now, go to the end of line to isolate it
            ;
            *Buffer = *LineStart
            
            While *Buffer\b And *Buffer\b <> 13 And *Buffer\b <> 10
              LineLength+1
              *Buffer+1
            Wend

            ; Note: we don't include the Chr(13) or Chr(10), it will be processed again in the next loop
            ;
            *Buffer = *LineStart ; Reset the buffer to the start of the line, and add the line length
            *Buffer + LineLength
            
            ; LogLine$ = Right(Filename$, Len(Filename$)-InitialPathLength-1)+": "+Str(Line)+" - "+PeekS(*LineStart, LineLength)
            LogLine$ = Filename$+": "+Str(Line)+" - "+ReplaceString(PeekS(*LineStart, LineLength), Chr(9), "  ") ; Replace all Tab with 2 spaces 
            
            AddGadgetItem(#GADGET_GrepOutput_List, -1, LogLine$)
            NbGrepFiles+1

          ElseIf *Buffer\b = 10  ; Unix files..
          
            Line+1
            *Buffer+1
            *LineStart = *Buffer
                     
          ElseIf *Buffer\b = 13 ; DOS compatible file line ending

            Line+1
            *Buffer+1
            
            If *Buffer\b = 10 ; Test is wrong, not a CRLF file
              *Buffer+1
            EndIf

            *LineStart = *Buffer
          Else
            *Buffer+1  ; Default case, just go on the next byte
          EndIf
        Wend
      
        FreeMemory(*BufferStart)
      EndIf 
    EndIf
    
    CloseFile(FileID)
  EndIf

EndProcedure


Procedure Grep(DirectoryID, Directory$, String$, PatternList$, InitialPathLength, Recursive, CaseSensitive)

  If ExamineDirectory(DirectoryID, Directory$, "")
  
    Repeat
      Type = NextDirectoryEntry()
            
      If Recursive And Type = 2 And DirectoryEntryName() <> ".." And DirectoryEntryName() <> "."

        Quit = Grep(DirectoryID+1, Directory$+DirectoryEntryName()+"\", String$, PatternList$, InitialPathLength, Recursive, CaseSensitive)
        UseDirectory(DirectoryID)
      
      ElseIf Type = 1
      
        Filename$  = DirectoryEntryName()
        Extension$ = LCase(GetExtensionPart(Filename$))
        
        
        If CheckExtension(PatternList$, Extension$)
          SearchStringInFile(#READFILE_SearchInFiles, Directory$+Filename$, String$, InitialPathLength, CaseSensitive)
        EndIf
            
      EndIf
      
      Quit = CheckGrepAbort() ; Handle the events to 'Stop'
            
    Until Type = 0 Or Quit 
  EndIf
  
  ProcedureReturn Quit
EndProcedure


Procedure ResizeGrepOutputWindow()

  UseWindow(#WINDOW_GrepOutput)
  Width  = WindowWidth()
  Height = WindowHeight()

  ResizeGadget(#GADGET_GrepOutput_List,   -1,  -1, Width-12, Height-10)
EndProcedure


Procedure OpenGrepOutputWindow()

  If IsGrepOutputWindow           ; If the window has been opened once, open it again (it allows to keep all parameters easily)
    HideWindow(#WINDOW_GrepOutput, #False)

  ElseIf OpenWindow(#WINDOW_GrepOutput, GrepOutputX, GrepOutputY, GrepOutputWidth, GrepOutputHeight, #PB_Window_SystemMenu | #PB_Window_Invisible | #PB_Window_SizeGadget, Langage(181), WindowID(#WINDOW_Main))

    CreateGadgetList(WindowID(#WINDOW_GrepOutput))
      ListViewGadget(#GADGET_GrepOutput_List, 6, 6, 0, 0)
      
    AddKeyboardShortcut(#WINDOW_GrepOutput, #PB_Shortcut_Escape, #GADGET_GrepOutput_Close)

    ResizeGrepOutputWindow()
       
    HideWindow(#WINDOW_GrepOutput, #False)
    
    IsGrepOutputWindow = #True
  EndIf
  
EndProcedure


Procedure GrepOutputWindowEvents(EventID)

  If EventID = #PB_Event_Menu     ; Little wrapper to map the shortcut events (identified as menu)
    EventID  = #PB_Event_Gadget   ; to normal gadget events...
    GadgetID = EventMenuID()
  Else
    GadgetID = EventGadgetID()
  EndIf

  Select EventID
    Case #PB_Event_CloseWindow
      Quit = 1
  
    Case #PB_Event_Gadget
      Select GadgetID
      
        Case #GADGET_GrepOutput_List
          If EventType() = #PB_EventType_LeftDoubleClick 
            Line$ = GetGadgetItemText(#GADGET_GrepOutput_List, GetGadgetState(#GADGET_GrepOutput_List), 0)

            Filename$  = StringField(Line$, 1, ":")
            If Len(Filename$) = 1 ; Probably only a letter of drive like 'C:'
              Filename$  = Filename$+":"+StringField(Line$, 2, ":")
            EndIf
            
            If Filename$ And Asc(Filename$) <> ';'  ; Not '; Search starting' or finished stuff
            
              Result = LoadSourceCodeReal(FileName$, 0)
              
              If Result
                LineNumber = Val(StringField(Right(Line$, Len(Line$)-Len(Filename$)-2), 1, " "))
                ChangeActiveLine(LineNumber, -3)
              EndIf
            EndIf
            
          EndIf
      
        Case #GADGET_GrepOutput_Close
          Quit = 1
      
      EndSelect
          
  EndSelect
  
  If Quit
    UseWindow(#WINDOW_GrepOutput)
    GrepOutputX      = WindowX()
    GrepOutputY      = WindowY()
    GrepOutputWidth  = WindowWidth()
    GrepOutputHeight = WindowHeight()

    HideWindow(#WINDOW_GrepOutput, #True)
    ClearGadgetItemList(#GADGET_GrepOutput_List)
    UseWindow(#WINDOW_Main)
  EndIf
  
EndProcedure


Procedure OpenGrepWindow()

  If IsGrepWindow  ; If the window has been opened once, open it again (it allows to keep all parameters easily)
    HideWindow(#WINDOW_Grep, #False)

  ElseIf OpenWindow(#WINDOW_Grep, 0, 0, #WINDOW_Grep_Width, #WINDOW_Grep_Height, #PB_Window_SystemMenu | #PB_Window_WindowCentered | #PB_Window_Invisible, Langage(172), WindowID(#WINDOW_Main))

    Left = 10 : Top = 10
    CreateGadgetList(WindowID(#WINDOW_Grep))
      TextGadget    (#GADGET_Grep_Label1  , Left, Top+2, 90                        , 18, Langage(173), #PB_Text_Right) : Left + 100  ; "Find :"
      ComboBoxGadget(#GADGET_Grep_FindWord, Left, Top  , #WINDOW_Grep_Width-Left-10, 400, #PB_ComboBox_Editable) : Left = 10 : Top + 30

      TextGadget    (#GADGET_Grep_Label2         , Left, Top+2, 90                        , 18, Langage(174), #PB_Text_Right) : Left + 100  ; "Find :"
      ComboBoxGadget(#GADGET_Grep_Directory      , Left, Top  , #WINDOW_Grep_Width-Left-45, 400, #PB_ComboBox_Editable|#CBS_AUTOHSCROLL) : Left +#WINDOW_Grep_Width-Left-40
      SendMessage_  (GadgetID(#GADGET_Grep_Directory), #CB_LIMITTEXT, 0, 0) ; remove text limit (#CBS_AUTOHSCROLL flag needs to be set on the gadget)
      
      ButtonGadget  (#GADGET_Grep_SelectDirectory, Left, Top  , 30, 21, "...") : Left = 10 : Top + 30
      
      TextGadget    (#GADGET_Grep_Label3 , Left, Top+2, 90                        , 18, Langage(175), #PB_Text_Right) : Left + 100  ; "Find :"
      ComboBoxGadget(#GADGET_Grep_Pattern, Left, Top  , #WINDOW_Grep_Width-Left-10, 400, #PB_ComboBox_Editable) : Left = 10 : Top + 30
      
      CheckBoxGadget(#GADGET_Grep_MatchCase, Left, Top+2, 140, 18, Langage(176)) : Left + #WINDOW_Grep_Width/2-10 ; "Match case"
      CheckBoxGadget(#GADGET_Grep_Recurse, Left, Top+2, 140, 18, Langage(177)) : Top + 35     ; "Whole word only"

      ButtonWidth = #WINDOW_Grep_Width/3
      
      ButtonGadget (#GADGET_Grep_Cancel    , ButtonWidth*2, Top, ButtonWidth-10, 25, Langage(180))  ; "Cancel"
      ButtonGadget (#GADGET_Grep_Stop      , ButtonWidth+5, Top, ButtonWidth-10, 25, Langage(179))  ; "Stop"
      ButtonGadget (#GADGET_Grep_Find      , 10           , Top, ButtonWidth-10, 25, Langage(178), #PB_Button_Default)  ; "Next"      

    AddKeyboardShortcut(#WINDOW_Grep, #PB_Shortcut_Return, #GADGET_Grep_Find)
    AddKeyboardShortcut(#WINDOW_Grep, #PB_Shortcut_Escape, #GADGET_Grep_Cancel)
    
    ; AddGadgetItem(#GADGET_Grep_Directory, -1,"test")
    ; SetGadgetText(#GADGET_Grep_Directory, "test")

    If GetGadgetText(#GADGET_Grep_Pattern) = ""
      AddGadgetItem(#GADGET_Grep_Pattern, -1,"*.pb")  ; Default to *.pb files, else it will find nothing
      SetGadgetState(#GADGET_Grep_Pattern, 0)
    EndIf

    ; AddGadgetItem(#GADGET_Grep_FindWord, -1,"while")
    ; SetGadgetState(#GADGET_Grep_FindWord, 0)
    
    DisableGrepWindow(#False)
    
    HideWindow(#WINDOW_Grep, 0)
    
    IsGrepWindow = #True
  EndIf
  
  SendMessage_(GadgetID(#GADGET_Grep_FindWord), #EM_SETSEL, 0, -1) ; Selects the whole text, so it's easily replaced
  ActivateGadget(#GADGET_Grep_FindWord)
  
EndProcedure


Procedure GrepWindowEvents(EventID)

  If EventID = #PB_Event_Menu     ; Little wrapper to map the shortcut events (identified as menu)
    EventID  = #PB_Event_Gadget   ; to normal gadget events...
    GadgetID = EventMenuID()
  Else
    GadgetID = EventGadgetID()
  EndIf

  Select EventID
    Case #PB_Event_CloseWindow
      Quit = 1
  
    Case #PB_Event_Gadget
      Select GadgetID
      
        Case #GADGET_Grep_SelectDirectory
          Directory$ = PathRequester("", GetGadgetText(#GADGET_Grep_Directory))
          If Directory$
            AddGadgetItem(#GADGET_Grep_Directory, -1,Directory$)
            SetGadgetText(#GADGET_Grep_Directory, Directory$)
          EndIf
      
        Case #GADGET_Grep_Find
          Directory$ = GetGadgetText(#GADGET_Grep_Directory)
          If Directory$
          
            If Right(Directory$, 1) <> "\" ; add it
              Directory$ + "\"
            EndIf
          
            String$ = GetGadgetText(#GADGET_Grep_FindWord)
            If String$
              DisableGrepWindow(#True)
              OpenGrepOutputWindow()
              
              NbGrepFiles = 0
              
              AddGadgetItem(#GADGET_GrepOutput_List, -1, "; "+Langage(192)+" '"+String$+"'...")
              
              If Grep(0, Directory$, String$, LCase(GetGadgetText(#GADGET_Grep_Pattern))+",", Len(Directory$), GetGadgetState(#GADGET_Grep_Recurse), GetGadgetState(#GADGET_Grep_MatchCase))
                AddGadgetItem(#GADGET_GrepOutput_List, -1, "; "+Langage(193)+" "+Str(NbGrepFiles)+" "+Langage(195))
              Else
                AddGadgetItem(#GADGET_GrepOutput_List, -1, "; "+Langage(194)+" "+Str(NbGrepFiles)+" "+Langage(195))
              EndIf
              
              DisableGrepWindow(#False)
              
              UpdateFindComboBox(#Gadget_Grep_FindWord)
              UpdateFindComboBox(#Gadget_Grep_Directory)
              UpdateFindComboBox(#Gadget_Grep_Pattern)
            Else
              MessageRequester(Langage(195), Langage(196), #MB_ICONINFORMATION)  ; "You have to enter a string to find."
              ActivateGadget(#GADGET_Grep_FindWord)
            EndIf
          Else
            MessageRequester(Langage(195), Langage(197), #MB_ICONINFORMATION)    ; "You have to choose a root path."
            ActivateGadget(#GADGET_Grep_Directory)
          EndIf
          
                    
        Case #GADGET_Grep_Cancel
          Quit = 1
      
      EndSelect
          
  EndSelect
  
  If Quit 
    HideWindow(#WINDOW_Grep, #True)
    UseWindow(#WINDOW_Main)
  EndIf
  
EndProcedure

; Test 
; 
; OpenWindow(0, 100, 100, 800, 400, #PB_Window_SystemMenu, "") : OpenGrepWindow() : 
; Repeat 
;   Event = WaitWindowEvent() 
;   If EventWindowID() = #WINDOW_Grep : GrepWindowEvents(Event)
;   ElseIf EventWindowID() = #WINDOW_GrepOutput : GrepOutputWindowEvents(Event) 
;   EndIf
; ForEver

; ExecutableFormat=Windows
; UseMainFile=PureBasic IDE.pb
; DisableDebugger
; EOF