;
; < -------------------- >
;   PureBasic IDE - Misc 
; < -------------------- >
;
; This program is free software; you can redistribute it and/or modify it under
; the terms of the GNU General Public License as published by the Free Software
; Foundation; either version 2 of the License, or (at your option) any later
; version. See the License.txt file for more details.
; 
; (c) 2003 Fantaisie Software
;

XIncludeFile "Common.pb"


Procedure.l LineNumbersStreamCallBack(dwCookie.l, pbBuff.l, cb.l, *pcb.l)
  
  If LineNumbersBufferLength = 0
    ProcedureReturn 1
  EndIf
  
  If LineNumbersBufferLength > cb
    CopyMemory(*LineNumbersBuffer, pbBuff, cb)
    PokeL(*pcb, cb)
    LineNumbersBufferLength-cb
    *LineNumbersBuffer+cb
  Else
    CopyMemory(*LineNumbersBuffer, pbBuff, LineNumbersBufferLength)
    PokeL(*pcb, LineNumbersBufferLength)
    LineNumbersBufferLength = 0
  EndIf
  
  ProcedureReturn 0
EndProcedure


Procedure UpdateLineNumbering()
  ;
  ; ATTENTION!
  ; 
  ; Im using a separate Callback for StreamIn here, and
  ; I dont use CopyMemoryString() because this stuff is
  ; NOT thread safe.
  ; The Highlighting stuff and this procedure are used
  ; within window callbacks (which are like 'threads'
  ; and are not reentrant safe), so its very important
  ; to completely separate it.
  ;
  ; Otherwise LOCKUPS are possible easily, especially
  ; because StreamCallback() uses Global variables
  ; for the copy-buffer.
  ;
  ; Dont change or 'optimize' this procedure to use
  ; StreamCallback() or CopyMemoryString() !!
  ;
  ; Danilo, 26.02.2004
  ;
  If *RichEditGadget And *LineNumbers And DisplayLineNumbers
    p.POINT
    p\x = 2
    p\y = 2
    If EditorFontSize
      p\y = EditorFontSize/2
    EndIf

    char  = SendMessage_(*RichEditGadget,#EM_CHARFROMPOS,0,@p)
    LineNumberStart = SendMessage_(*RichEditGadget,#EM_EXLINEFROMCHAR,0,char) + 1

    GetClientRect_(*RichEditGadget,rect.RECT)
    p\y = rect\bottom - p\y
    char = SendMessage_(*RichEditGadget,#EM_CHARFROMPOS,0,@p)
    LineNumberEnd = SendMessage_(*RichEditGadget,#EM_EXLINEFROMCHAR,0,char) + 1

    LineCount = SendMessage_(*RichEditGadget,#EM_GETLINECOUNT,0,0)
    
    If LineCount =< LineNumberEnd
      LineNumberEnd = LineCount
    Else
      LineNumberEnd + 1
    EndIf

    If LineNumberStart <> OldLineNumberStart Or LineNumberEnd <> OldLineNumberEnd
      
      OldLineNumberStart = LineNumberStart
      OldLineNumberEnd   = LineNumberEnd

      length = (Len(StrU(LineNumberEnd,#LONG))+6)*(LineNumberEnd-LineNumberStart+5)+2000

      buffer = AllocateMemory(length)
      If buffer
        mem = buffer
        ;acp.l = GetACP_()
        ;If acp=0 : acp=1252 : EndIf ; Default if none available?
        CopyMemory(@"{\rtf1\ansi\ansicpg1252\uc1 \deff0{\colortbl;",mem,45) : mem + 45

        Color$ = "\red"+Str(LineNumberColor & $FF)+"\green"+Str((LineNumberColor >> 8) & $FF)+"\blue"+Str((LineNumberColor >> 16) & $FF)+";"
        CopyMemory(@Color$,mem,Len(Color$)) : mem + Len(Color$)

        If EditorFontStyle & #PB_FONT_BOLD
          CopyMemory(@"}\b\cf1\qr ",mem,11) : mem + 11
        Else
          CopyMemory(@"}\b0\cf1\qr ",mem,12) : mem + 12
        EndIf

        For a = LineNumberStart To LineNumberEnd
          A$  = StrU(a,#LONG)+"\line "
          len = Len(A$)
          CopyMemory(@A$,mem,len) : mem + len
        Next a

        CopyMemory(@"}",mem,1) : mem + 1

        SendMessage_ (*LineNumbers,#EM_HIDESELECTION,1,0)

        range.CHARRANGE
        range\cpMin =  0
        range\cpMax = -1
        SendMessage_(*LineNumbers,#EM_SETSEL,range\cpMin,range\cpMax)

        *LineNumbersBuffer      = buffer
        LineNumbersBufferLength = mem - buffer
        StreamStructure.EDITSTREAM\pfnCallback = @LineNumbersStreamCallBack()
        SendMessage_(*LineNumbers,#EM_STREAMIN,#SF_RTF,@StreamStructure)

        SendMessage_ (*LineNumbers,#EM_HIDESELECTION,1,0)

        FreeMemory(buffer)
        
        ; check/set size after streamin new lines
        For a = 1 To Len(StrU(LineCount,#LONG))
          WidthChar$ + "9"
        Next a
        
        hDC = GetDC_(*LineNumbers)
        If hDC
          hOldFont = SelectObject_(hDC,*LineNumberFontHandle)
          If GetTextExtentPoint32_(hDC,@WidthChar$,Len(WidthChar$),@width)
            width + 7
            If LineNumberWidth <> width
              LineNumberWidth = width
              SendMessage_(*RichEditGadget,#WM_SETREDRAW,0,0)
              SendMessage_(*LineNumbers   ,#WM_SETREDRAW,0,0)
              ResizeRichEditGadget()
              SendMessage_(*RichEditGadget,#WM_SETREDRAW,1,0)
              SendMessage_(*LineNumbers   ,#WM_SETREDRAW,1,0)
              InvalidateRect_(*RichEditGadget,0,1)
              InvalidateRect_(*LineNumbers   ,0,1)
              UpdateWindow_(*RichEditGadget)
              UpdateWindow_(*LineNumbers)
            EndIf
          EndIf
          
          SelectObject_(hDC, hOldFont)
          ReleaseDC_(*LineNumbers, hDC)
        EndIf
      EndIf
    EndIf

  EndIf
EndProcedure


Procedure LineNumbersCallBack(hWnd,Msg,wParam,lParam)
  If Msg = #WM_SETFOCUS
    If *RichEditGadget
      SetFocus_(*RichEditGadget)
    EndIf
  ElseIf Msg = #WM_LBUTTONDOWN Or Msg = #WM_LBUTTONUP Or Msg = #WM_LBUTTONDBLCLK
    UpdateLineNumbering()
  ElseIf Msg = #WM_RBUTTONDOWN Or Msg = #WM_RBUTTONUP Or Msg = #WM_RBUTTONDBLCLK
  ElseIf Msg = #WM_MBUTTONDOWN Or Msg = #WM_MBUTTONUP Or Msg = #WM_MBUTTONDBLCLK
  ElseIf Msg = #WM_CAPTURECHANGED Or Msg = #WM_MOUSEMOVE Or Msg = #WM_MOUSEWHEEL
  Else
    Result = CallWindowProc_(*LineNumberOldCallback,hWnd,Msg,wParam,lParam)
  EndIf
  SetCursor_(LineNumberCursor)
  ProcedureReturn Result
EndProcedure


Procedure CreateLineNumbers()
  ; #WS_EX_CLIENTEDGE
  *LineNumbers = CreateWindowEx_(0,"RichEdit20A",0,#ES_READONLY|#ES_NOHIDESEL|#WS_CHILD|#WS_VISIBLE|#ES_MULTILINE,0,0,0,0,WindowID(#WINDOW_Main),#GADGET_LineNumbers,GetModuleHandle_(0),0)
  If *LineNumbers
    LineNumberCursor = LoadCursor_(0,#IDC_ARROW)
    *LineNumberOldCallback = SetWindowLong_(*LineNumbers,#GWL_WNDPROC,@LineNumbersCallBack())
    SendMessage_(*LineNumbers,#EM_SETMARGINS,#EC_LEFTMARGIN,3)
    ;#ENM_SCROLLEVENTS = $00000008
    ;SendMessage_(*LineNumbers,#EM_SETEVENTMASK ,0,#ENM_PROTECTED)
    SendMessage_(*LineNumbers,#EM_SETBKGNDCOLOR,0,LineNumberBackColor)
  EndIf
EndProcedure



; ExecutableFormat=Windows
; EOF