;
; < -------------------- >
;   PureBasic IDE - Main
; < -------------------- >
;
; This program is free software; you can redistribute it and/or modify it under
; the terms of the GNU General Public License as published by the Free Software
; Foundation; either version 2 of the License, or (at your option) any later
; version. See the License.txt file for more details.
;
; (c) 2003 Fantaisie Software
;
; Note: When compiling, makes sure than your editor doesn't have the
;       'Run only once instance' flag set, else it will quit immediately..
;
;         

Declare  SaveSourceCodeAs()
Declare  SaveSourceCode(FileName$, Flags)
Declare$ GetCurrentWord()
Declare  ResizeStructureViewerWindow()
Declare  ResizeGrepOutputWindow()
Declare  UpdateProcedureList()
Declare  OpenProcedureBrowser()
Declare  RichEdit_InsertString(String.s, Update)
Declare  LoadSourceCodeReal(FileName$, ForceReload)
Declare  CreateIDEMenu(DisableRedraw)
Declare  SwitchSource(State)
Declare  UpdatePanelFilename()
Declare  UpdateToolTips()
Declare  UpdateLineNumbering()
Declare  UpdateProcedureList()
Declare  ResizeRichEditGadget() 

XIncludeFile "Common.pb"
XIncludeFile "Misc.pb"
XIncludeFile "Syntax Highlighting.pb"
XIncludeFile "Menu_AddHelpFiles.pb"
XIncludeFile "AddTools.pb"
XIncludeFile "RecentFiles.pb"
XIncludeFile "RunOnce.pb"
XIncludeFile "SelectionComments.pb"
XIncludeFile "RichEdit_OpenIncludeFile.pb"
XIncludeFile "CheckPureBasicKeyWords.pb"
XIncludeFile "StructureViewer.pb"
XIncludeFile "DisplayPlatformSDKHelp.pb"
XIncludeFile "LineNumbers.pb"

Procedure MAKELPARAM(low, high)
  ProcedureReturn low + (high<<16)
EndProcedure


Procedure SelectionMenus(selection)
  ; Disables or Enables MenuItems, depending
  ; on the selection in the RichEdit window
  ;
  ; selection = 1: there is something selected
  ; selection = 0: there is nothing selected
  If selection <> RichEdit_Selection
    RichEdit_Selection = selection
    If selection
      state = #MF_ENABLED
    Else
      state = #MF_GRAYED
    EndIf
    EnableMenuItem_(*MainMenu,          #MENU_Cut                ,#MF_BYCOMMAND|state)
    EnableMenuItem_(*MainMenu,          #MENU_Copy               ,#MF_BYCOMMAND|state)
    EnableMenuItem_(*MainMenu,          #MENU_CommentSelection   ,#MF_BYCOMMAND|state)
    EnableMenuItem_(*MainMenu,          #MENU_UnCommentSelection ,#MF_BYCOMMAND|state)
    
    EnableMenuItem_(*RichEdit_PopupMenu,#MENU_Cut                ,#MF_BYCOMMAND|state)
    EnableMenuItem_(*RichEdit_PopupMenu,#MENU_Copy               ,#MF_BYCOMMAND|state)
    EnableMenuItem_(*RichEdit_PopupMenu,#MENU_CommentSelection   ,#MF_BYCOMMAND|state)
    EnableMenuItem_(*RichEdit_PopupMenu,#MENU_UnCommentSelection ,#MF_BYCOMMAND|state)
  EndIf
EndProcedure


Procedure CheckForRichEditSelection()
  ; Check if there is a current selection
  ; in the RichEdit window
  If *RichEditGadget
    SendMessage_(*RichEditGadget,#EM_EXGETSEL,0,selrange.CHARRANGE)
    If selrange\cpMin <> selrange\cpMax
      RichEdit_Selection = 0
      SelectionMenus(1)
    Else
      RichEdit_Selection = 1
      SelectionMenus(0)
    EndIf
  EndIf
EndProcedure


Procedure EditorKeyHandler(VirtualKey)
  Shared VK_ReturnState
  
  If VirtualKey = #VK_RETURN And VK_ReturnState = 0
    
    PreviousCursorX = CursorPositionX-1
    
    ; First, delete all the text which could be highlighted (default notepad feature)
    ;
    SendMessage_(*RichEditGadget, #EM_REPLACESEL, 1, "")

    SyntaxHighlighting()
    
    *Line = AllocateMemory(2000)
    PokeW(*Line, 1000)
    
    SendMessage_(*RichEditGadget, #EM_GETSEL        , @StartPosition, 0)
    StartPosition = SendMessage_(*RichEditGadget, #EM_EXLINEFROMCHAR, 0             , StartPosition)
    SendMessage_(*RichEditGadget, #EM_GETLINE       , StartPosition , *Line)   ; Get the current line buffer..
    
    VK_ReturnState = 1 ; Very tricky. This inform the WindowCallback to ignore the following #VK_RETURN return, else it's a dead loop
    
    SendMessage_(*RichEditGadget, #WM_KEYDOWN, #VK_RETURN, 1)
    
    ClearEffects() ; Reset all the font effects for the newline..

     *Cursor.Byte = *Line
    While *Cursor\b = ' ' And PreviousCursorX > 0   ; Insert space at the start of the newline (auto-indent)
      *Cursor+1
      PreviousCursorX-1
      NbSpaces+1
    Wend
    
    If NbSpaces
      SendMessage_(*RichEditGadget, #EM_REPLACESEL, 1, Space(NbSpaces))
    EndIf
    
    FreeMemory(*Line)
    
  ElseIf VirtualKey = #VK_TAB
    
    If GetKeyState_(#VK_SHIFT) & $8000
      ShiftMode = 1
    Else
      ShiftMode = 0
    EndIf
    
    SendMessage_(*RichEditGadget, #EM_GETSEL        , @StartPosition, @EndPosition)
    FirstLineIndex   = SendMessage_(*RichEditGadget, #EM_EXLINEFROMCHAR, 0             , StartPosition)
    LastLineIndex    = SendMessage_(*RichEditGadget, #EM_EXLINEFROMCHAR, 0             , EndPosition)
    LastLinePosition = SendMessage_(*RichEditGadget, #EM_LINEINDEX     , LastLineIndex , 0)
        
    *Buffer = AllocateMemory(EndPosition-StartPosition+1)
    
    If LastLinePosition = EndPosition And LastLineIndex <> FirstLineIndex ; Needed to handle the case where the selection is at the very end of the
      LastLineIndex-1                                                      ; line and then the new line is take in count too.
    EndIf
    
    If FirstLineIndex <> LastLineIndex ; Block indent when pressing TAB
      
      For k=FirstLineIndex To LastLineIndex
        
        LineIndex  = SendMessage_(*RichEditGadget, #EM_LINEINDEX , k        , 0)
        LineLength = SendMessage_(*RichEditGadget, #EM_LINELENGTH, LineIndex, 0)
        
        If LineLength > 0 And OldLineIndex+LineLength+1 <> EndPosition
          
          If ShiftMode
            If TabLength > LineLength
              RealTabLength = LineLength
            Else
              RealTabLength = TabLength
            EndIf
            
            SendMessage_(*RichEditGadget, #EM_SETSEL    , LineIndex, LineIndex+RealTabLength)
            SendMessage_(*RichEditGadget, #EM_GETSELTEXT, 0        , *Buffer)
            
            Counter = RealTabLength
            *Cursor = *Buffer
            While (*Cursor\b = 9 Or *Cursor\b = ' ') And Counter
              *Cursor+1
              Counter-1
            Wend
            
            If Counter < RealTabLength
              SendMessage_(*RichEditGadget, #EM_SETSEL, LineIndex, LineIndex+RealTabLength-Counter)
              SendMessage_(*RichEditGadget, #EM_REPLACESEL, 1, "")
            EndIf
            
            Offset-(RealTabLength-Counter)
          Else
            SendMessage_(*RichEditGadget, #EM_SETSEL, LineIndex, LineIndex) ; Cursor to the beginning of the line
            
            ClearEffects() ; Reset all the font effects for the newline..
            
            SendMessage_(*RichEditGadget, #EM_REPLACESEL, 1, Space(TabLength))
            
            Offset+TabLength
          EndIf
        EndIf
      Next
      
      SendMessage_(*RichEditGadget, #EM_SETSEL, StartPosition, EndPosition+Offset)
    Else
      ClearEffects()  ; Reset all the font effects for start of the newline..
      
      RealTabLength = TabLength-((CursorPositionX-1) % TabLength)  ; Smart indent
      
      SendMessage_(*RichEditGadget, #EM_REPLACESEL, 1, Space(RealTabLength))
    EndIf
    
    FreeMemory(*Buffer)    
  Else
    VK_ReturnState = 0
  EndIf
  
EndProcedure


Procedure SetEditorFont(Font.s, Size)

  If LoadFont(#FONT_Editor, Font, Size, EditorFontStyle)
    *LineNumberFontHandle = UseFont(#FONT_Editor)
    
    SendMessage_(*RichEditGadget, #WM_SETFONT, *LineNumberFontHandle, MAKELPARAM(1, 0))
    
    If *LineNumbers
      SendMessage_(*LineNumbers,#WM_SETFONT,*LineNumberFontHandle,MAKELPARAM(1, 0))
      UpdateLineNumbering()
    EndIf
    
  EndIf

EndProcedure


Procedure.l GetToolbarButtonState(ButtonID)
  
  debuginfo.TBBUTTONINFOA\cbSize = SizeOf(TBBUTTONINFOA)
  debuginfo\dwMask = #TBIF_STATE
  If SendMessage_(*MainToolBar, #TB_GETBUTTONINFOA, ButtonID, @debuginfo)
    If debuginfo\fsState & #TBSTATE_CHECKED
      State = 1
    EndIf
  EndIf
  
  ProcedureReturn State

EndProcedure


Procedure SetToolbarButtonState(ButtonID, State)

  debuginfo.TBBUTTONINFOA\cbSize = SizeOf(TBBUTTONINFOA)
  debuginfo\dwMask = #TBIF_STATE
  If SendMessage_(*MainToolBar, #TB_GETBUTTONINFOA, ButtonID, @debuginfo)
    If State : State = 1 : EndIf
    debuginfo\fsState = (debuginfo\fsState & ~#TBSTATE_CHECKED) | (State * #TBSTATE_CHECKED)
    debuginfo\dwMask = #TBIF_STATE
    SendMessage_(*MainToolBar, #TB_SETBUTTONINFOA, ButtonID, @debuginfo)
  EndIf

EndProcedure


Procedure UpdateDebuggerState()

  CompilerIf #Demo = 0
    If CountList(FileList()) 
      SetMenuItemState  (0, #MENU_Debugger, FileList()\Debugger)
      SetToolbarButtonState(#MENU_Debugger, FileList()\Debugger)
    EndIf
  CompilerEndIf

EndProcedure


Procedure CreateIDEMenu(DisableRedraw)
  
  If DisableRedraw
    SendMessage_(*MainWindow, #WM_SETREDRAW, 0, 0)  ; Needed to remove any flickering when recreating the menu on runtime
  EndIf
  
  *MainMenu = CreateMenu(0, WindowID(#WINDOW_Main))
  If *MainMenu
    
    ; ---------------------------------- File
    MenuTitle(Langage(0))
    MenuItem(#MENU_New   , Langage(1))
    MenuItem(#MENU_Open  , Langage(2))
    MenuItem(#MENU_Save  , Langage(3))
    MenuItem(#MENU_SaveAs, Langage(4))
    MenuItem(#MENU_Close , Langage(104))
    MenuBar();
    ; MenuItem(#MENU_Print, Langage(5))
    MenuItem(#MENU_Preference, Langage(37))
    Menu_AddRecentFiles()
    MenuBar()
    MenuItem(#MENU_Exit, Langage(6))
    
    ; ---------------------------------- Edition
    MenuTitle(Langage(7));
    MenuItem(#MENU_Undo, Langage(67))
    MenuItem(#MENU_Redo, Langage(68))
    MenuBar()
    MenuItem(#MENU_Cut  , Langage(8))
    MenuItem(#MENU_Copy , Langage(9))
    MenuItem(#MENU_Paste, Langage(10))
    MenuBar()
    MenuItem(#MENU_CommentSelection,   Langage(142))
    MenuItem(#MENU_UnCommentSelection, Langage(143))
    MenuBar()
    MenuItem(#MENU_SelectAll, Langage(105))
    MenuBar()
    MenuItem(#MENU_Goto, Langage(36))
    MenuBar()
    MenuItem(#MENU_Find, Langage(11))
    MenuItem(#MENU_Next, Langage(12))
    MenuItem(#MENU_FindInFiles, Langage(171))
    
    ; ---------------------------------- Compiler
    MenuTitle(Langage(13))
    MenuItem(#MENU_CompileRun      , Langage(14))
    MenuItem(#MENU_Run             , Langage(15))
    MenuBar()
    MenuItem(#MENU_Debugger   , Langage(16))
    MenuBar()
    MenuItem(#MENU_RestartCompiler , Langage(187))
    MenuBar()
    MenuItem(#MENU_CompilerOption  , Langage(17))
    MenuItem(#MENU_CreateExecutable, Langage(18))
    
    ; ---------------------------------- Tools
    MenuTitle(Langage(141))
    MenuItem(#MENU_VisualDesigner , Langage(153))        
    MenuItem(#MENU_ASCIITable     , Langage(106))
    MenuItem(#MENU_StructureViewer, Langage(107))
    MenuItem(#MENU_PickColor      , Langage(108))
    MenuItem(#MENU_SmartUpdate    , Langage(169))    
    MenuBar()
    MenuItem(#MENU_AddTools       , Langage(109))
    AddTools_AddMenuEntries()
    
    ; ---------------------------------- Help
    MenuTitle(Langage(19))
    MenuItem(#MENU_Help , Langage(20))
    Menu_AddHelpFiles()
    MenuItem(#MENU_About, Langage(21))
    
  EndIf
  
  CompilerIf #Demo = 1
    SetMenuItemState  (0, #MENU_Debugger, 1)
    SetToolbarButtonState(#MENU_Debugger, 1)
    DisableMenuItem      (#MENU_Debugger, 1)
    DisableToolBarButton (#MENU_Debugger, 1)
  CompilerEndIf

  UpdateDebuggerState()
  
  If DisableRedraw
    SendMessage_(*MainWindow, #WM_SETREDRAW, 1, 0)
    InvalidateRect_(*MainWindow, 0, 0)
  EndIf
  
EndProcedure


Procedure CreateIDEPopupMenu()
  *RichEdit_PopupMenu = CreatePopupMenu(1)
  If *RichEdit_PopupMenu
    MenuItem(#MENU_Cut  , Langage(8))
    MenuItem(#MENU_Copy , Langage(9))
    MenuItem(#MENU_Paste, Langage(10))
    MenuBar()
    MenuItem(#MENU_CommentSelection,   Langage(142))
    MenuItem(#MENU_UnCommentSelection, Langage(143))
    MenuBar()
    MenuItem(#MENU_SelectAll, Langage(105))
    
    If AddHelpFiles_Counter
      Menu_AddHelpFiles()
    Else
      MenuBar()
    EndIf
    MenuItem(#MENU_Help  , Langage(20))
    MenuItem(#MENU_Close , Langage(104))
  EndIf
EndProcedure


Procedure ChangeStatus(Text$)

  StatusBarText(0, 0, Str(CursorPositionX)+":"+Str(CursorPositionY), 4)
  
  If Text$
    StatusBarText(0, 1, Text$, 2)
  EndIf
EndProcedure


Procedure UpdateStatusBar2()
  Shared PreviousYValue, PreviousXValue, PreviousQuickHelp$
  
  SendMessage_(*RichEditGadget, #EM_EXGETSEL, 1, Range.CHARRANGE)
  
  CurrentLineIndex = SendMessage_(*RichEditGadget, #EM_EXLINEFROMCHAR, 0, Range\cpMin)
  CursorPositionY = CurrentLineIndex+1
  
  CurrentLineStartIndex = SendMessage_(*RichEditGadget, #EM_LINEINDEX, CurrentLineIndex, 0)
  CursorPositionX = Range\cpMax-CurrentLineStartIndex+1
  
  If QuickHelpPosition > -1
    If IsQuickHelpWinAPI = 1
      QuickHelp$ + PeekS(*APIFunctionsArray(QuickHelpPosition, 0))
      
      If *APIFunctionsArray(QuickHelpPosition, 1)
        QuickHelp$ + "_" + PeekS(*APIFunctionsArray(QuickHelpPosition, 1))
      Else
        QuickHelp$ + "_ - "+Langage(189)
      EndIf    
    Else
      QuickHelp$ + PeekS(*BasicFunctionsArray(QuickHelpPosition, 0))
      
      If *BasicFunctionsArray(QuickHelpPosition, 1)
        QuickHelp$ + " " + PeekS(*BasicFunctionsArray(QuickHelpPosition, 1))
      EndIf
    EndIf
  EndIf
    
  If QuickHelp$ And QuickHelp$ = PreviousQuickHelp$  ; Don't change the status for a flicker free status bar update (visible on XP !)
    QuickHelp$ = ""
  EndIf
  
  ; Test needed to have a flicker free status bar display...
  
  If PreviousYValue <> CurrentLineIndex Or PreviousXValue <> Range\cpMax-CurrentLineStartIndex+1 Or QuickHelp$
    ChangeStatus(QuickHelp$)
  EndIf
  
  PreviousYValue     = CurrentLineIndex
  PreviousXValue     = Range\cpMax-CurrentLineStartIndex+1
  
  If QuickHelp$
    PreviousQuickHelp$ = QuickHelp$  ; To keep track of the statusbar text (to remove flickering)
  EndIf
 
EndProcedure


Procedure ResizeRichEditGadget()

  If ShowMainToolbar
    MainToolbarHeight = 28
  Else
    MainToolbarHeight = 0
  EndIf
  
  If DisplayLineNumbers
    LineNr = LineNumberWidth
  Else
    LineNr = 0
  EndIf

  If *RichEditGadget
    GetClientRect_(*MainWindow, WindowSize.RECT)
    ;BorderX = GetSystemMetrics_(#SM_CXEDGE)
    ;BorderY = GetSystemMetrics_(#SM_CYEDGE)
    ;ResizeGadget(#GADGET_EditFrame, EditLeftOffset, MainToolbarHeight+EditTopOffset+23, WindowSize\right-EditRightOffset-EditLeftOffset, WindowSize\bottom-20-MainToolbarHeight-EditBottomOffset-23)    
    
    MoveWindow_(*RichEditGadget,EditLeftOffset+LineNr+BorderX,  MainToolbarHeight+EditTopOffset+23+BorderY, WindowSize\right-EditRightOffset-EditLeftOffset-LineNr-BorderX*2, WindowSize\bottom-20-MainToolbarHeight-EditBottomOffset-23-BorderY*2, 1)
    ; BringWindowToTop_(*RichEditGadget) ; set on top of Frame3dGadget
    If *LineNumbers
      MoveWindow_(*LineNumbers   ,EditLeftOffset+BorderX       ,  MainToolbarHeight+EditTopOffset+23+BorderY, LineNr                                                , WindowSize\bottom-20-MainToolbarHeight-EditBottomOffset-23-BorderY*2, 1)
      ; BringWindowToTop_(*LineNumbers) ; set on top of Frame3dGadget
    EndIf
           
  EndIf
  
EndProcedure


Procedure WindowCallBack(Window, Message, wParam, lParam)
  Shared VK_ReturnState, HighlightIT
  ; Static NeedForcedRedraw
  
  ReturnValue = #PB_ProcessPureBasicEvents 
    
  If Window = *MainWindow

    If Message = #WM_Size
      
      GetClientRect_(*MainWindow, WindowSize.RECT)

      If ShowMainToolbar
        MainToolbarHeight = 28
      Else
        MainToolbarHeight = 0
      EndIf
      
      If *ProcedureBrowser

        ResizeGadget(#GADGET_FilesPanel, EditLeftOffset, MainToolbarHeight, WindowSize\right-EditRightOffset, 22)

        If wParam = #SIZE_MAXIMIZED ;Or wParam = #SIZE_RESTORED
          ResizeRichEditGadget()
          UpdateLineNumbering()
        Else
          ResizeRichEditGadget()
        EndIf

        If ProcedureBrowserType
          If ProcedureBrowserSide ; Left
            MoveWindow_(*ProcedureBrowser,0,MainToolbarHeight+EditTopOffset, EditLeftOffset-3, WindowSize\bottom-20-MainToolbarHeight-EditBottomOffset, 1)
          Else                    ; Right
            MoveWindow_(*ProcedureBrowser,WindowSize\right-EditRightOffset+3,MainToolbarHeight+EditTopOffset, EditRightOffset-3, WindowSize\bottom-20-MainToolbarHeight-EditBottomOffset, 1)
          EndIf
        EndIf

      EndIf

      If hCloseImage
        ResizeGadget(#GADGET_ToolbarCloseButton,WindowSize\right-17,5,-1,-1)
      EndIf
      
      ReturnValue = 0

    ElseIf Message = #WM_EXITSIZEMOVE

      UpdateLineNumbering() ; update line numbers after resize

    ElseIf Message = #WM_Activate       ; For automatic RichEdit activation when the editor get the focus..
      
      If wParam & $FFFF <> #WA_INACTIVE
        UseWindow(#WINDOW_Main)
        SetFocus_(*RichEditGadget)
        ReturnValue = 0
      EndIf
    
    ElseIf Message = #WM_Command
    
      If wParam >> 16 = #EN_CHANGE And wParam & $FFFF = #GADGET_RichText And NoRichEditChanges = #False

        If HighlightIT = 2

          ; This Line MUST BE BEFORE SyntaxHighlighting(), otherwise
          ; an endless loop will happen on Win9x and editor crashes on key press !!
          ;
          HighlightIT = 0 
          SyntaxHighlighting()
        EndIf
        
        ; Fix the last redraw bug, when line changes. It need to be redrawed at the last pass,
        ; so we implemented a counter system as we don't really know when (it has 4 pass max,
        ; so redraw at each pass)
        ;
        ; Very ugly hack, but hey... if you find a cleaner way to remove flickering from RichEdit
        ; when doing syntax highlightning, just do it ! :)
        ;
        ;If NeedForcedRedraw > 0
        ;  NeedForcedRedraw - 1
        ;  InvalidateRect_(*RichEditGadget, 0, 1)
        ;EndIf
        
        UpdatePanelFilename()
        UpdateStatusBar2()
        
        LineCount = SendMessage_(*RichEditGadget,#EM_GETLINECOUNT,0,0)
        If CurrentLineNumberCount <> LineCount
          CurrentLineNumberCount = LineCount
          UpdateLineNumbering()
        ;  NeedForcedRedraw = 4   ; Needed to fix a redraw bug when pressing ENTER at the bottom of the sources
        EndIf
        
      ElseIf wParam >> 16 = #CBN_CLOSEUP
        
        SetFocus_(*RichEditGadget)
      
      ElseIf wParam >> 16 = #EN_VSCROLL

        UpdateLineNumbering() ; update line numbers after vertical scrolling

      EndIf
      
    ElseIf Message = #WM_Notify
      
      *MSGFilter.MSGFILTER = lParam
      
      
      If *MSGFilter\nmhdr\hwndFrom = *RichEditGadget  ; A notification message from the RichEdit gadget

        If *MSGFilter\nmhdr\code = #EN_SELCHANGE
          *CurSel.SELCHANGE = lParam
          If *CurSel\chrg\cpMin <> *CurSel\chrg\cpMax
            SelectionMenus(1) ; There is a Selection
          Else
            SelectionMenus(0) ; There is no Selection
          EndIf
        ElseIf *MSGFilter\msg = #WM_KEYDOWN
          
          Virtualkey = *MSGFilter\wParam
          
          If Virtualkey  = #VK_UP Or Virtualkey = #VK_DOWN Or Virtualkey = #VK_RIGHT Or Virtualkey = #VK_LEFT Or Virtualkey = #VK_NEXT Or Virtualkey = #VK_PRIOR Or Virtualkey = #VK_HOME Or Virtualkey = #VK_END Or Virtualkey  = #VK_SPACE
            If HighlightIT
              HighlightIT = 0
              SyntaxHighlighting()
            EndIf
          EndIf
          
          If *MSGFilter\wParam = #VK_TAB Or (*MSGFilter\wParam = #VK_RETURN And VK_ReturnState = 0)
            EditorKeyHandler(*MSGFilter\wParam)
            ProcedureReturn -1
          Else
            VK_ReturnState = 0
          EndIf
          
        ElseIf *MSGFilter\msg = #WM_KEYUP
          
          Virtualkey = *MSGFilter\wParam
          
          If Virtualkey = #VK_BACK Or Virtualkey = #VK_CLEAR Or Virtualkey = #VK_DELETE
            SyntaxHighlighting()
            UpdateProcedureList()
            
          ElseIf Virtualkey  = #VK_UP Or Virtualkey = #VK_DOWN Or Virtualkey = #VK_RETURN
            UpdateProcedureList()
          EndIf
          
        ElseIf *MSGFilter\msg = #WM_CHAR
          
          Key = *MSGFilter\wParam & $FF ; make it unsigned byte only
          
          If ValidCharacters(Key)
            HighlightIT = 1 + RealTimeSyntaxHighlighting
            
          ElseIf Key = ':' Or Key = 34 Or Key = ' ' Or Key = '#' Or Key = '$' Or Key = '=' Or Key = '-' Or Key = '+' Or Key = '/' Or Key = '\' Or Key = '*' Or Key = '>' Or Key = '<' Or Key = '.' Or Key = ',' Or Key = ';' Or Key = '(' Or Key = ')' Or Key = '[' Or Key = ']' Or Key = 39  Or Key = '%' Or Key = '&' Or Key = '|' Or Key = '!' Or Key = '~' Or Key = '%' Or Key = '?' Or Key = '@'
            HighlightIT = 2
          EndIf
          
          If *MSGFilter\wParam = 9 ; TAB only...
            *MSGFilter\wParam = 0  ; Replace the original TAB message with a null one... It removes the TAB
            ReturnValue = 1        ; Return a non null value tell to the handler to process our new message
          EndIf

        EndIf
        
        UpdateStatusBar2()

      EndIf
      
    ElseIf Message = #WM_CTLCOLORLISTBOX And lParam = *ProcedureBrowser
      SetBkMode_(wParam, #TRANSPARENT)
      SetTextColor_(wParam,ProcedureBrowserFrontColor)
      ReturnValue = ProcedureBrowserBackBrush
      
    ElseIf Message = Editor_RunOnceMessage
      If PeekS(wParam,9) = "OPENFILE"
        RunOnce_Message()
        ReturnValue = 0
      EndIf
    EndIf

  ElseIf IsStructureViewerWindow And Window = WindowID(#WINDOW_StructureViewer)

    If Message = #WM_Size
      ResizeStructureViewerWindow()
      ReturnValue = 0
      
    ElseIf Message = #WM_CTLCOLORLISTBOX And lParam = GadgetID(#GADGET_StructureViewer_List)
    
      SetBkMode_(wParam, #TRANSPARENT)
      SetTextColor_(wParam,ProcedureBrowserFrontColor)
      ReturnValue = ProcedureBrowserBackBrush    
    EndIf

  ElseIf Message = #WM_Size And IsGrepOutputWindow And Window = WindowID(#WINDOW_GrepOutput)
    ResizeGrepOutputWindow()
    ReturnValue = 0
  EndIf

  ProcedureReturn ReturnValue
  
EndProcedure


Procedure ShowToolbar(Show)
  If *MainToolBar
    If Show
      ShowWindow_(*MainToolBar,#SW_SHOWNA)
    Else
      ShowWindow_(*MainToolBar,#SW_HIDE)
    EndIf
  EndIf
  SendMessage_(WindowID(#WINDOW_Main),#WM_SIZE,0,0)
EndProcedure


Procedure CreateGUI()
  
  If OpenWindow(#WINDOW_Main, EditorWindowX , EditorWindowY, EditorWindowWidth, EditorWindowHeight, #PB_Window_Invisible | #PB_Window_SystemMenu | #PB_Window_SizeGadget | #PB_Window_MaximizeGadget | #PB_Window_MinimizeGadget, "PureBasic")
    SetWindowCallback(@WindowCallBack())
    
    ; To remove big flickering but makes lot of refresh problem (Toolbar, panel etc..)
    ;
    ; SetWindowLong_(WindowID(), #GWL_STYLE, GetWindowLong_(WindowID(), #GWL_STYLE) | #WS_CLIPCHILDREN)

    RemoveKeyboardShortcut(#WINDOW_Main,-1)
    
    *MainWindow = WindowID(#WINDOW_Main)
    
    CreateGadgetList(*MainWindow)        
    CreateIDEMenu(0)
    DisableMenuItem(#MENU_StructureViewer, 1)
   
    ; Turn on the drag & drop feature..
    ;
    DragAcceptFiles_(WindowID(#WINDOW_Main), 1)
    
    ; Create our panel gadget for multi file support
    ;
    UseGadgetList(WindowID(#WINDOW_Main))
    PanelGadget(#GADGET_FilesPanel, 0, 0, 0, 40)
    
    ; Create Frame3D for 3d border around the edit area
    ;
    ;Frame3DGadget(#GADGET_EditFrame, 0, 0, 0, 0, "", #PB_Frame3D_Double)
    
    ;ResizeWindow(1600, 1200)  ; a 1600 width toolbar... Quick hack, should be changed.
    
    *MainToolBar = CreateToolBar(0, WindowID(#WINDOW_Main))
    If *MainToolBar
      ToolBarImageButton(#MENU_New,  UseImage(CatchImage(#PB_Any, ?NewIcon)))
      ToolBarImageButton(#MENU_Open, UseImage(CatchImage(#PB_Any, ?OpenIcon)))
      ToolBarImageButton(#MENU_Save, UseImage(CatchImage(#PB_Any, ?SaveIcon)))
      ToolBarSeparator()
      ToolBarImageButton(#MENU_Close,UseImage(CatchImage(#PB_Any, ?FileCloseIcon)))
      ToolBarSeparator()
      ToolBarImageButton(#MENU_Cut,   UseImage(CatchImage(#PB_Any, ?CutIcon)))
      ToolBarImageButton(#MENU_Copy,  UseImage(CatchImage(#PB_Any, ?CopyIcon)))
      ToolBarImageButton(#MENU_Paste, UseImage(CatchImage(#PB_Any, ?PasteIcon)))
      ToolBarSeparator()
      ToolBarImageButton(#MENU_Undo, UseImage(CatchImage(#PB_Any, ?UndoIcon)))
      ToolBarImageButton(#MENU_Redo, UseImage(CatchImage(#PB_Any, ?RedoIcon)))
      ToolBarSeparator()
      ToolBarImageButton(#MENU_VisualDesigner, UseImage(CatchImage(#PB_Any,?VisualDesignerIcon)))
      ToolBarSeparator()
      ToolBarImageButton(#MENU_Debugger,   UseImage(CatchImage(#PB_Any, ?DebuggerIcon)))
      ToolBarImageButton(#MENU_CompileRun, UseImage(CatchImage(#PB_Any, ?CompileRunIcon)))
      
      ; Make the debugger icon a toggle button, to mimic the behaviour of the menu item
      ;
      debuginfo.TBBUTTONINFOA\cbSize = SizeOf(TBBUTTONINFOA)
      debuginfo\dwMask = #TBIF_STYLE
      If SendMessage_(*MainToolBar, #TB_GETBUTTONINFOA, #MENU_Debugger, @debuginfo)
        debuginfo\fsStyle = debuginfo\fsStyle | #TBSTYLE_CHECK
        debuginfo\dwMask = #TBIF_STYLE
        SendMessage_(*MainToolBar, #TB_SETBUTTONINFOA, #MENU_Debugger, @debuginfo)
      EndIf
      
      UpdateToolTips()
      
      DisableToolBarButton (#MENU_CompileRun, 1)
      DisableMenuItem(#MENU_CompileRun, 1)
    EndIf
    ShowToolbar(ShowMainToolbar)

    ;UseGadgetList(*MainToolbar)
    ;  CreateCloseImage()
    ;  If hCloseImage
    ;    ButtonImageGadget(#GADGET_ToolbarCloseButton,1500,5,16,14,hCloseImage)
    ;  EndIf
    UseGadgetList(*MainWindow)

    OpenProcedureBrowser()
    
    CreateIDEPopupMenu()
    
    *StatusBar = CreateStatusBar(0, *MainWindow)
    If *StatusBar
      AddStatusBarField(80)
      AddStatusBarField(6000)
      CursorPositionX = 1 : CursorPositionY = 1
      ChangeStatus("Welcome to PureBasic !")
      GetWindowRect_(*StatusBar, StatusBarRect.RECT)
      EditBottomOffset = (StatusBarRect\bottom-StatusBarRect\top)-20 ; On XP, the statusbar height isn't the same
    EndIf
    
    CreateLineNumbers()
    
    ResizeWindow(EditorWindowWidth, EditorWindowHeight)  
  EndIf
  
EndProcedure


Procedure ChangeActiveLine(NewLine, Offset)
  NewLine-1
  NewLineIndex = SendMessage_(*RichEditGadget, #EM_LINEINDEX, NewLine, 0)
  SendMessage_(*RichEditGadget, #EM_SETSEL, NewLineIndex, NewLineIndex)
  SendMessage_(*RichEditGadget, #EM_LINESCROLL, 0, -20000)
  SendMessage_(*RichEditGadget, #EM_LINESCROLL, 0, NewLine+Offset)
EndProcedure



XIncludeFile "PreferenceWindow.pb"
XIncludeFile "FindWindow.pb"
XIncludeFile "GotoWindow.pb"
XIncludeFile "OptionWindow.pb"
XIncludeFile "AboutWindow.pb"
XIncludeFile "GrepWindow.pb"
XIncludeFile "ASCIITable.pb"
XIncludeFile "ProcedureParser.pb"
XIncludeFile "PickColor.pb"



Procedure SourceCodeSaved()
  
  If SendMessage_(*RichEditGadget, #EM_GETMODIFY, 0, 0) ; Check is the *RichEditGadget has been modified
    
    If FileList()\FileName$ = ""
      FileName$ = "<New>"
    Else
      FileName$ = GetFilePart(FileList()\FileName$)
    EndIf
    
    Select MessageRequester("PureBasic", FileName$+Chr(13)+Langage(34)+Chr(10)+Langage(35), #MB_YESNOCANCEL | #MB_ICONWARNING)
      Case #IDYES
        Result = SaveSourceCode(FileList()\FileName$, 0)
        
      Case #IDNO
        Result = 1
        
      Default
        Result = 0
        
    EndSelect
  Else
    Result = 1
  EndIf
  
  ProcedureReturn Result
EndProcedure


Procedure RichEditCallback(Window,Message,wParam,lParam)
  
  ;Debug Message
  
  Select Message
      ; left doubleclick - special actions
      ; (default = select current word)
    Case #WM_LBUTTONDBLCLK
      A$ = UCase(GetCurrentWord())
      If A$ = "XINCLUDEFILE" Or A$ = "INCLUDEFILE"
        RichEdit_OpenIncludeFile()
        returnvalue = 0
      ElseIf ProcedureBrowser_JumpOnCodeDoubleClick() ; returns 1 if successfully jumped to procedure
        returnvalue = 0
      Else
        returnvalue = CallWindowProc_(FileList()\RichEdit_OldHandler,Window,Message,wParam,lParam)
      EndIf

    ; fix for RichEdit resize flickering, dont remove!
    Case #WM_WINDOWPOSCHANGING
      returnvalue = 0
    
    ; virtualize mousewheel, needed for LineNumbering
    Case #WM_MOUSEWHEEL
      If wParam & $FFFF = 0
        If (wParam >> 16) > 0
          x = -1
        ElseIf (wParam >> 16) < 0
          x =  1
        EndIf
        If x
          If UseCustomScroll
            lines_to_scroll = CustomScrollLines
          Else
            If SystemParametersInfo_(#SPI_GETWHEELSCROLLLINES, 0, @lines_to_scroll, 0) = 0
              lines_to_scroll = 3
            EndIf
          EndIf
          For a = 1 To lines_to_scroll
            SendMessage_(*RichEditGadget,#EM_LINESCROLL,0,x)
          Next
        EndIf
      EndIf
      returnvalue = 0

    ; check vertical scroll events, needed for LineNumbering
    Case #WM_VSCROLL
      ; #EM_GETFIRSTVISIBLELINE,0,0
      ; #EM_SCROLL,#SB_LINEDOWN,0
      If wParam & $FFFF = #SB_LINEDOWN
        SendMessage_(*RichEditGadget,#EM_LINESCROLL,0, 1)
      ElseIf wParam & $FFFF = #SB_LINEUP
        SendMessage_(*RichEditGadget,#EM_LINESCROLL,0,-1)
      ElseIf wParam & $FFFF = #SB_PAGEUP Or wParam & $FFFF = #SB_PAGEDOWN
        GetClientRect_(*RichEditGadget,rect.RECT)
        p.POINT\x = 2 : p\y = 2
        char = SendMessage_(*RichEditGadget,#EM_CHARFROMPOS,0,p.POINT)
        line  = SendMessage_(*RichEditGadget,#EM_EXLINEFROMCHAR,0,char)
        p\y = rect\bottom - 2
        char = SendMessage_(*RichEditGadget,#EM_CHARFROMPOS,0,p.POINT)
        line = SendMessage_(*RichEditGadget,#EM_EXLINEFROMCHAR,0,char) - line
        If wParam & $FFFF = #SB_PAGEUP
          SendMessage_(*RichEditGadget,#EM_LINESCROLL,0,-line)
        Else
          SendMessage_(*RichEditGadget,#EM_LINESCROLL,0, line)
        EndIf
      Else
        returnvalue = CallWindowProc_(FileList()\RichEdit_OldHandler,Window,Message,wParam,lParam)
      EndIf


      If wParam & $FFFF = #SB_ENDSCROLL ; Ends scroll.
        If OldScrollEvent <> #SB_LINEUP And OldScrollEvent <> #SB_LINEDOWN And OldScrollEvent <> #SB_PAGEDOWN And OldScrollEvent <> #SB_PAGEUP
        
          ; Disable the richedit redraw, else the #EM_LINESCROLL -1 followed by #EM_LINESCROLL 1
          ; Produce a bad flickering artifact
          ;
          SendMessage_(*RichEditGadget, #WM_SETREDRAW, 0, 0)
        
          p.POINT
          p\x = 2
          p\y = 2
          If EditorFontSize
            p\y = EditorFontSize/2
          EndIf
          char = SendMessage_(*RichEditGadget,#EM_CHARFROMPOS,0,p.POINT)
          line  = SendMessage_(*RichEditGadget,#EM_EXLINEFROMCHAR,0,char)
          If line > 0
            linecount = SendMessage_(*RichEditGadget,#EM_GETLINECOUNT,0,0)
            GetClientRect_(*RichEditGadget,rect.RECT)
            p\y = rect\bottom - p\y
            char = SendMessage_(*RichEditGadget,#EM_CHARFROMPOS,0,p.POINT)
            line = SendMessage_(*RichEditGadget,#EM_EXLINEFROMCHAR,0,char)

            If line+1 <> linecount
              SendMessage_(*RichEditGadget,#EM_LINESCROLL,0,-1)
              SendMessage_(*RichEditGadget,#EM_LINESCROLL,0, 1)
            Else
              SendMessage_(*RichEditGadget,#EM_LINESCROLL,0, -1) ; If the source is scrolled with the slider at the end
              SendMessage_(*RichEditGadget,#EM_LINESCROLL,0, 2)  ; and line number is activated, then the linenumber and text are misaligned. This code fix that.
            EndIf
          Else
            SendMessage_(*RichEditGadget,#EM_LINESCROLL,0, 1)
            SendMessage_(*RichEditGadget,#EM_LINESCROLL,0,-1)
          EndIf
          
          ; Enable the automatic redraw again, and refresh the whole richedit window
          ;
          SendMessage_(*RichEditGadget,#WM_SETREDRAW, 1, 0)
          InvalidateRect_(*RichEditGadget, 0, 0)
        EndIf
      EndIf
      
      UpdateLineNumbering()
      OldScrollEvent = wParam & $FFFF

    Default
      returnvalue = CallWindowProc_(FileList()\RichEdit_OldHandler,Window,Message,wParam,lParam)
  EndSelect
  ProcedureReturn returnvalue
EndProcedure


Procedure CreateNewPanelItem(FileName$)
  
  If *RichEditGadget
    ShowWindow_(*RichEditGadget, #SW_HIDE)
  EndIf
  
  ; #WS_EX_CLIENTEDGE ; removed for LineNumbers, otherwise it has a border
  *RichEditGadget = CreateWindowEx_(0, "RichEdit20A", 0, #ES_NOHIDESEL | #WS_CHILD | #WS_VISIBLE | #ES_MULTILINE | #ES_AUTOHSCROLL | #ES_AUTOVSCROLL | #ES_WANTRETURN | #WS_HSCROLL | #WS_VSCROLL, 0, 0, 0, 0, WindowID(#Window_Main), #GADGET_RichText, GetModuleHandle_(0), 0)

  ; Try to find out if the locale uses an ANSI codepage by default
  linfo.s = Space(6)
  linfo_len.l = GetLocaleInfo_(GetSystemDefaultLCID_(), #LOCALE_IDEFAULTANSICODEPAGE, @linfo, 6)
  If linfo_len>0
    ; If default system ansi code page is not zero, use the ANSI code page
    ;
    If Val(Mid(linfo, 1, linfo_len))<>0
     SendMessage_(*RichEditGadget, #EM_SETTEXTMODE, #TM_SINGLECODEPAGE, 0)
    EndIf
  EndIf
  
  SendMessage_(*RichEditGadget, #EM_SETMARGINS,#EC_LEFTMARGIN,3) 

  ResizeRichEditGadget()
  
  SetEditorFont(EditorFontName, EditorFontSize)
  
  #ENM_SCROLLEVENTS = $00000008
  SendMessage_(*RichEditGadget, #EM_SETEVENTMASK , 0, #ENM_CHANGE | #ENM_SELCHANGE | #ENM_KEYEVENTS | #ENM_MOUSEEVENTS | #ENM_SCROLLEVENTS | #ENM_SCROLL | #ENM_PROTECTED )
  SendMessage_(*RichEditGadget, #EM_SETBKGNDCOLOR, 0, BackgroundColor)
  SendMessage_(*RichEditGadget, #EM_EXLIMITTEXT  , 0, $7FFFFFFF) ; Put the capacity to the max !
  
  ClearEffects()
  ClearGadgetItemList(#GADGET_ProcedureList)
  ClearGadgetItemList(#GADGET_ProcedureBrowser)
  ClearList( Procedure_Browser_OldList() )
  
  LastElement(FileList())
  AddElement(FileList())
  FileList()\RichEdit  = *RichEditGadget
  FileList()\RichEdit_OldHandler = SetWindowLong_(*RichEditGadget,#GWL_WNDPROC,@RichEditCallback())
  FileList()\FileName$ = FileName$
  FileList()\Debugger  = 1

  UpdateDebuggerState()
  
  If FileName$
    Title$ = GetFilePart(FileName$)
  Else
    Title$ = "<New>"
  EndIf
  
  SetWindowText_(*MainWindow, "PureBasic - "+Title$)

  AddGadgetItem(#GADGET_FilesPanel, -1, Title$)
  SetGadgetState(#GADGET_FilesPanel, CountGadgetItems(#GADGET_FilesPanel)-1)
  CheckForRichEditSelection()

  UpdateLineNumbering()

EndProcedure


Procedure NewSourceCode()
  
  CreateNewPanelItem("")
  
  FileList()\ExecutableFormat$  = "Windows"
  
  QuickHelpPosition = -1;
  SendMessage_(*RichEditGadget, #WM_SETTEXT, 0, "")
  SendMessage_(*RichEditGadget, #EM_SETMODIFY, 0, 0)
  ClearEffects()
  SetFocus_(*RichEditGadget)
  
  UpdateStatusBar2()
  
  ProcedureBrowserMustUpdate = 1
  UpdateProcedureList()
  
  InvalidateRect_(*MainWindow,0,0)
  
EndProcedure


Procedure AnalyseProjectSettings(*Buffer, Length)
  
  *Cursor.Byte   = *Buffer+Length
  *EndTextBuffer = *Cursor
  *LastPointer   = *Cursor  ; Length can be 0, so we have to initialize it here too
  
  While (Length)
    
    *LastPointer = *Cursor
    
    While (Length And *Cursor\b <> 10)
      Length-1
      *Cursor-1
    Wend
    
    If Length
      Line$ = PeekS(*Cursor+1)
      
      If Line$ = "; EOF"
      ElseIf Line$ =          "; EnableNT4"         : FileList()\EnableNT4 = 1
      ElseIf Line$ =          "; EnableAsm"         : FileList()\EnableASM = 1
      ElseIf Line$ =          "; EnableXP"          : FileList()\EnableXP  = 1
      ElseIf Line$ =          "; EnableOnError"     : FileList()\EnableOnError = 1
      ElseIf Line$ =          "; DisableDebugger"   : FileList()\Debugger  = 0 :
      ElseIf Left(Line$,13) = "; Executable="       : FileList()\ExecutableName$   = Right(Line$, Len(Line$)-13)
      ElseIf Left(Line$,19) = "; ExecutableFormat=" : FileList()\ExecutableFormat$ = Right(Line$, Len(Line$)-19)
      ElseIf Left(Line$,14) = "; CommandLine="      : FileList()\CommandLine$      = Right(Line$, Len(Line$)-14)
      ElseIf Left(Line$,10) = "; UseIcon="          : FileList()\UseIcon = 1     : FileList()\IconName$ = Right(Line$, Len(Line$)-10)
      ElseIf Left(Line$,14) = "; UseMainFile="      : FileList()\UseMainFile = 1 : FileList()\MainFile$ = Right(Line$, Len(Line$)-14)
      ElseIf Left(Line$,6)  = "; CPU="              : FileList()\CPU          = Val(Right(Line$, Len(Line$)-6 ))
      ElseIf Left(line$,17) = "; CursorPosition="   : FileList()\CursorPosition = Val(Right(line$, Len(line$)-17))
      ElseIf Left(line$,12) = "; FirstLine="        : FileList()\FirstLine      = Val(Right(line$, Len(line$)-12))
      Else : Length = 0
      EndIf
      
      If Length
        *Cursor-1
        *Cursor\b = 0
        *Cursor-1
      EndIf
    EndIf
  Wend

  If *EndTextBuffer <> *LastPointer ; At least one option has been found, so we have to put back the carriage return
    PokeB(*LastPointer+1, 10)
    PokeB(*LastPointer+2, 0)
  EndIf
  
EndProcedure



Procedure.l StreamCallBack(dwCookie.l, pbBuff.l, cb.l, *pcb.l)
  
  If CallbackBufferLength = 0
    ProcedureReturn 1
  EndIf
  
  If CallbackBufferLength > cb
    CopyMemory(*CallbackBuffer, pbBuff, cb)
    PokeL(*pcb, cb)
    CallbackBufferLength-cb
    *CallbackBuffer+cb
  Else
    CopyMemory(*CallbackBuffer, pbBuff, CallbackBufferLength)
    PokeL(*pcb, CallbackBufferLength)
    CallbackBufferLength = 0
  EndIf
  
  ProcedureReturn 0
EndProcedure


Procedure LoadSourceCodeReal(FileName$, ForceReload)
  
  UseWindow(#WINDOW_Main)
  
  ; Check if the same source is already opened
  ;
  If ForceReload = 0 And CountList(FileList())
    CurrentFile = @FileList()
  
    ForEach FileList()
      If UCase(FileList()\FileName$) = UCase(FileName$)
        AlreadyOpened = 1
        Index = ListIndex(FileList())
        Break
      EndIf
    Next
    
    ChangeCurrentElement(FileList(), CurrentFile)
  EndIf
  
  
  If AlreadyOpened

    SetGadgetState(#GADGET_FilesPanel, Index) 
    SwitchSource(Index)
    Result = 1
  
  ElseIf ReadFile(#READFILE_LOADSOURCECODEREAL, FileName$)
    
    If SendMessage_(*RichEditGadget, #WM_GETTEXTLENGTH, 0, 0) = 0 And SendMessage_(*RichEditGadget, #EM_GETMODIFY, 0, 0) = 0
      ;SetGadgetText(#GADGET_FilesPanel, GetFilePart(FileName$))
      FileList()\FileName$ = FileName$
      UpdatePanelFilename()
    Else
      CreateNewPanelItem(FileName$)
    EndIf
    SetWindowText_(*MainWindow, "PureBasic - "+FileName$)
    
    Menu_AddToRecentFiles(FileName$)
    
    ChangeStatus(Langage(26)) ; "Loading source code..."
    
    FileLength = Lof()
    TextBuffer = LocalAlloc_(#LMEM_ZEROINIT, FileLength+10)
    
    ReadData(TextBuffer, FileLength)
    CloseFile(#READFILE_LOADSOURCECODEREAL)
    
    AnalyseProjectSettings(TextBuffer, FileLength)

    UpdateDebuggerState()

    UpdateSyntaxHighlightning = 1
    
    UpdateTextHighlightning(EnableColoring, TextBuffer, FileLength)
    
    SendMessage_(*RichEditGadget, #EM_SETSEL, 0, 0)
             
    If FileList()\FirstLine And MemorizeCursor ; Memorize Cursorposition
      ChangeActiveLine(FileList()\FirstLine,0)
      NewLineIndex = SendMessage_(*RichEditGadget, #EM_LINEINDEX, FileList()\CursorPosition, 0)
      SendMessage_(*RichEditGadget, #EM_SETSEL, NewLineIndex, NewLineIndex)
    EndIf

    SendMessage_(*RichEditGadget, #EM_SETMODIFY, 0, 0)
    SetFocus_(*RichEditGadget)

    LocalFree_(TextBuffer)
    ChangeStatus(Langage(27))  ; "File loaded"
    
    ProcedureBrowserMustUpdate = 1
    UpdateProcedureList()
    
    ; Check Selection in RichEdit window
    CheckForRichEditSelection()
    
    ;SetGadgetText(#GADGET_FilesPanel, GetFilePart(FileName$))
    ;SetGadgetText(#GADGET_FilesPanel
    UpdatePanelFilename()
    
    OldLineNumberStart = -1 ; force update
    UpdateLineNumbering()
    
    Result = 1
  Else
    MessageRequester(Langage(32), Langage(28)+FileName$, #MB_OK | #MB_ICONERROR)  ; "Impossible to read the file !"
  EndIf
  
  ProcedureReturn Result
EndProcedure


Procedure LoadSourceCode()
  
  If FileList()\FileName$ = ""
    Path$ = SourceDirectory
  Else
    Path$ = FileList()\FileName$
  EndIf
  
  File$ = OpenFileRequester("", Path$, "PureBasic Source (*.pb)|*.PB|Assembly Source (*.asm,*.s)|*.asm;*.s|All (*.*)|*.*", 0, #PB_Requester_MultiSelection)
  If File$ <> ""
    Repeat
      LoadSourceCodeReal(File$, 0)
      
      File$ = NextSelectedFileName()
    Until File$ = ""
  EndIf
EndProcedure


Procedure SaveSourceCode(FileName$, Flags)
  
  If FileName$
    If CreateFile(#CREATEFILE_SAVESOURCECODE, FileName$)
      ChangeStatus(Langage(29)) ; "Saving source code..."
      
      FileLength = SendMessage_(*RichEditGadget, #WM_GETTEXTLENGTH, 0, 0)
      TextBuffer = LocalAlloc_(#LMEM_ZEROINIT, FileLength+1)
      
      SendMessage_(*RichEditGadget, #WM_GETTEXT, FileLength+1, TextBuffer)
      
      WriteData(TextBuffer, FileLength)
      LocalFree_(TextBuffer)
      
      WriteStringN("")
      WriteStringN("; ExecutableFormat="+FileList()\ExecutableFormat$)
            
      FileList()\FirstLine = SendMessage_(*RichEditGadget, #EM_GETFIRSTVISIBLELINE , 0, 0) + 1
      FileList()\CursorPosition = CursorPositionY - 1
      
      If MemorizeCursor
        If FileList()\CursorPosition  : WriteStringN("; CursorPosition=" + Str(FileList()\CursorPosition)) : EndIf
        If FileList()\FirstLine       : WriteStringN("; FirstLine="      + Str(FileList()\FirstLine))      : EndIf
      EndIf
      
      If FileList()\EnableASM       : WriteStringN("; EnableAsm")                                        : EndIf
      If FileList()\EnableNT4       : WriteStringN("; EnableNT4")                                        : EndIf
      If FileList()\EnableXP        : WriteStringN("; EnableXP")                                         : EndIf
      If FileList()\EnableOnError   : WriteStringN("; EnableOnError")                                    : EndIf
      If FileList()\UseIcon         : WriteStringN("; UseIcon="        + FileList()\IconName$)           : EndIf
      If FileList()\UseMainFile     : WriteStringN("; UseMainFile="    + FileList()\MainFile$)           : EndIf
      If FileList()\ExecutableName$ : WriteStringN("; Executable="     + FileList()\ExecutableName$)     : EndIf
      If FileList()\CPU             : WriteStringN("; CPU="            + Str(FileList()\CPU))            : EndIf
      If FileList()\CommandLine$    : WriteStringN("; CommandLine="    + FileList()\CommandLine$)        : EndIf
      If FileList()\Debugger = 0    : WriteStringN("; DisableDebugger")                                  : EndIf
      
      WriteString("; EOF")
      CloseFile(#CREATEFILE_SAVESOURCECODE)
      
      ChangeStatus(Langage(30)) ; "Source code saved."
      
      If Flags = 0
        SendMessage_(*RichEditGadget, #EM_SETMODIFY, 0, 0)
        Menu_AddToRecentFiles(FileName$)  
      EndIf
      
      If FileName$ <> FullPath + "Compilers\PB_EditorOutput.pb" ; Changes the title only if it's not the temporary file name
        SetWindowText_(*MainWindow, "PureBasic - "+FileName$)
      EndIf
      
      ;SetGadgetText(#GADGET_FilesPanel, GetFilePart(FileName$))
      UpdatePanelFilename()
      
      Result = 1
    Else
      MessageRequester(Langage(32), Langage(31)+FileName$, #MB_OK | #MB_ICONERROR)  ; "Can't save the file..."
    EndIf
  Else
    Result = SaveSourceCodeAs()
  EndIf
  
  ProcedureReturn Result
EndProcedure


Procedure SaveSourceCodeAs()
  
  If FileList()\FileName$ = ""
    Path$ = SourceDirectory
  Else
    Path$ = FileList()\FileName$
  EndIf
  
  File$ = SaveFileRequester("", Path$, "PureBasic Source (*.pb)|*.PB|All (*.*)|*.*", 0)
  If File$
    Result = 1;
    
    If SelectedFilePattern() = 0 And UCase(Right(File$,3)) <> ".PB"
      File$+".pb"
    EndIf
    
    If FileSize(File$) > -1
      Result = MessageRequester("PureBasic", "The file '"+File$+"' already exists."+Chr(10)+"Do you want to overwrite it ?", #MB_YESNO | #MB_ICONWARNING)
      If Result <> #IDYES
        Result = 0
      EndIf
    EndIf
  EndIf
  
  If Result
    ;SetGadgetText(#GADGET_FilesPanel, GetFilePart(File$))
    FileList()\FileName$ = File$
    UpdatePanelFilename()
    
    Result = SaveSourceCode(File$, 0)
  EndIf
  
  ProcedureReturn Result
EndProcedure

#EM_REDO = (#WM_USER+84)

Procedure Undo()
  ;While SendMessage_(*RichEditGadget, #EM_GETUNDONAME, 0, 0) = 0 And SendMessage_(*RichEditGadget, #EM_CANUNDO, 0, 0)
  ;  SendMessage_(*RichEditGadget, #EM_UNDO, 0, 0)
  ;Wend
  
  SendMessage_(*RichEditGadget, #EM_UNDO, 0, 0)

  UpdateLineNumbering()
  UpdateProcedureList()
EndProcedure


Procedure Redo()
  ;While SendMessage_(*RichEditGadget, #EM_GETREDONAME, 0, 0) = 0 And SendMessage_(*RichEditGadget, #EM_CANREDO, 0, 0)
  ;  SendMessage_(*RichEditGadget, #EM_REDO, 0, 0)
  ;Wend
  
  SendMessage_(*RichEditGadget, #EM_REDO, 0, 0)

  UpdateLineNumbering()
  UpdateProcedureList()
EndProcedure


Procedure ClipboardCut()
  SendMessage_(*RichEditGadget, #WM_CUT, 0, 0)

  ;ProcedureBrowserMustUpdate = 1
  UpdateLineNumbering()
  UpdateProcedureList()
EndProcedure


Procedure ClipboardCopy()
  SendMessage_(*RichEditGadget, #WM_COPY, 0, 0)
EndProcedure


Procedure ClipboardPaste()
  
  If EnableColoring
    
    If OpenClipboard_(0)
      
      ClipboardText = GetClipboardData_(#CF_TEXT)
      If ClipboardText
        
        TextLength = MemoryStringLength(ClipboardText)
        
        ; Get the start of the line, before the cursor..
        
        SendMessage_(*RichEditGadget, #EM_REPLACESEL   , 1, "")  ; Always replace the current selection with nothing..
        SendMessage_(*RichEditGadget, #EM_HIDESELECTION, 1, 0)
        SendMessage_(*RichEditGadget, #EM_EXGETSEL     , 1, Range.CHARRANGE)
        
        NeededLine    = SendMessage_(*RichEditGadget, #EM_EXLINEFROMCHAR, 1         , Range\cpMin)
        StartPosition = SendMessage_(*RichEditGadget, #EM_LINEINDEX     , NeededLine, 0)
        
        SendMessage_(*RichEditGadget, #EM_SETSEL, StartPosition, Range\cpMin) ; Select only the start of the line
        
        TextBuffer = AllocateMemory(TextLength+1+Range\cpMin-StartPosition)
        If TextBuffer
          SendMessage_(*RichEditGadget, #EM_GETSELTEXT, 0, TextBuffer)
          
          CopyMemory(ClipboardText, TextBuffer+Range\cpMin-StartPosition, TextLength+1)
          
          *RTFBuffer = BuildRTFBuffer(TextBuffer, MemoryStringLength(TextBuffer))
          If (*RTFBuffer)
            *CallbackBuffer       = *RTFBuffer;
            CallbackBufferLength = MemoryStringLength(*CallbackBuffer)
            
            StreamStructure.EDITSTREAM\pfnCallback = @StreamCallBack()  ; One line is enough :)
            
            SendMessage_(*RichEditGadget, #EM_STREAMIN, #SF_RTF | #SFF_SELECTION, StreamStructure)
            GlobalFree_ (*RTFBuffer)
          EndIf
          
          FreeMemory(TextBuffer)
        EndIf
        
        SendMessage_(*RichEditGadget, #EM_HIDESELECTION, 0, 0)
      EndIf
      
      CloseClipboard_()
    EndIf
  Else
    SendMessage_(*RichEditGadget, #EM_PASTESPECIAL, #CF_TEXT, 0)
  EndIf
  
  ;ProcedureBrowserMustUpdate = 1
  UpdateLineNumbering()
  UpdateProcedureList()
  
EndProcedure


Procedure RichEdit_InsertString(String.s, Update)
  ;
  ; Inserts a String in the current RichEdit Control
  ; Used by "Pick Color" in Toolsmenu... dont remove
  ; Can also be used by other procedures.
  ;
  ; Also used by Structure Viewer beginning with Editor PB v3.80
  ;
  
  ; Get the start of the line, before the cursor..
  
  SendMessage_(*RichEditGadget, #EM_REPLACESEL   , 1, "")  ; Always replace the current selection with nothing..
  SendMessage_(*RichEditGadget, #EM_HIDESELECTION, 1, 0)
  
  If EnableColoring
    TextLength = Len(String)

    SendMessage_(*RichEditGadget, #EM_EXGETSEL     , 1, Range.CHARRANGE)
    
    NeededLine    = SendMessage_(*RichEditGadget, #EM_EXLINEFROMCHAR, 1         , Range\cpMin)
    StartPosition = SendMessage_(*RichEditGadget, #EM_LINEINDEX     , NeededLine, 0)

    SendMessage_(*RichEditGadget, #EM_SETSEL, StartPosition, Range\cpMin) ; Select only the start of the line
    
    TextBuffer = AllocateMemory(TextLength+1+Range\cpMin-StartPosition)
    If TextBuffer
      SendMessage_(*RichEditGadget, #EM_GETSELTEXT, 0, TextBuffer)
      
      CopyMemory(@String, TextBuffer+Range\cpMin-StartPosition, TextLength+1)
      
      *RTFBuffer = BuildRTFBuffer(TextBuffer, MemoryStringLength(TextBuffer))
      If (*RTFBuffer)
        *CallbackBuffer       = *RTFBuffer;
        CallbackBufferLength = MemoryStringLength(*CallbackBuffer)
        
        StreamStructure.EDITSTREAM\pfnCallback = @StreamCallBack()  ; One line is enough :)
        
        SendMessage_(*RichEditGadget, #EM_STREAMIN, #SF_RTF | #SFF_SELECTION, StreamStructure)
        GlobalFree_ (*RTFBuffer)
      EndIf
      FreeMemory(TextBuffer)
    EndIf
    
  Else
    SendMessage_(*RichEditGadget, #EM_REPLACESEL, #TRUE, String)
  EndIf
  
  SendMessage_(*RichEditGadget, #EM_HIDESELECTION, 0, 0)

  ; Update is optional as when we use it with CommentBlock it will be ugly to update it at each line
  ;
  If Update
    UpdateLineNumbering()
    UpdateProcedureList()
  EndIf

EndProcedure


Procedure RestartCompiler()

  If IsProgramRunning = 1 ; make sure, the compiler isn't busy!
    MessageRequester("PureBasic",Langage(188), #MB_ICONWARNING)             
    
  Else

    If IsStructureViewerWindow <> 0 
      CloseStructureViewerWindow()
    EndIf
    
    TerminateProcess_(CompilerProcessID,0) 
    
    ClearList(StructureList$()) ; make sure, the Structureslist gets updated
    
    CompilerProcessID = 0
    CompilerThreadID = 0
 
    DisableToolBarButton(#MENU_CompileRun, 1)
    DisableMenuItem(#MENU_CompileRun, 1)
    DisableMenuItem(#MENU_StructureViewer, 1)       

    Parameters.s = Str(GetCurrentThreadId_())+" /STANDBY"
    CompilerProcessID = RunProgram(FullPath+"Compilers\PBCompiler.exe", Parameters, "", 2)    
    
  EndIf
  
EndProcedure


Procedure DisplayCompilingWindow()
  
  WindowWidth  = 200
  WindowHeight =  40
  
  If OpenWindow(#WINDOW_Compiling, 0, 0, WindowWidth, WindowHeight, #PB_Window_WindowCentered, "PureBasic", WindowID(#WINDOW_Main))
    CreateGadgetList(WindowID(#WINDOW_Compiling))
      TextGadget(0, 10, 10, 180, 20, Langage(24))
    
    While WindowEvent() : Wend ; Be sure than the message queue is empty, else the compiler message could be lost (Traumatic bug)
  EndIf
  
EndProcedure


Procedure HideCompilingWindow()
  CloseWindow(#WINDOW_Compiling)
  UseWindow(#WINDOW_Main)
EndProcedure


Procedure BuildCompilerFlags()
  
  If FileList()\EnableASM     : Result | #PB_FLG_INLINEASM     : EndIf
  If FileList()\EnableNT4     : Result | #PB_FLG_ENABLENT4     : EndIf
  If FileList()\EnableXP      : Result | #PB_FLG_ENABLEXP      : EndIf
  If FileList()\EnableOnError : Result | #PB_FLG_ENABLEONERROR : EndIf
  If FileList()\Debugger      : Result | #PB_FLG_DEBUGGER      : EndIf
  
  If FileList()\ExecutableFormat$ = "Console"    : Result | #PB_FLG_CONSOLE : EndIf
  If FileList()\ExecutableFormat$ = "Shared Dll" : Result | #PB_FLG_DLL     : EndIf
  
  Select FileList()\CPU
    Case 1
      Result | #PB_FLG_CPU_DYNAMIC
    Case 2
      Result | #PB_FLG_CPU_MMX
    Case 3
      Result | #PB_FLG_CPU_3DNOW
    Case 4
      Result | #PB_FLG_CPU_SSE
    Case 5
      Result | #PB_FLG_CPU_SSE2
  EndSelect
  
  ProcedureReturn Result
EndProcedure


Procedure CompileRun()
  Shared SharedFileName, IsProgramRunning
  
  If CompilerThreadID
    
    DisplayCompilingWindow()

    ; First, select the good filename if a main file is used
    ;
    If FileList()\UseMainFile
      Filename$ = FileList()\MainFile$
      
      If GetPathPart(FileName$) = ""
        Filename$ = GetPathPart(FileList()\FileName$)+Filename$
      EndIf
    Else
      Filename$ = FileList()\FileName$
    EndIf
    
    If AutoSave And FileList()\FileName$ And SendMessage_(*RichEditGadget, #EM_GETMODIFY, 0, 0)   ; <New> filename doesn't need to be automatically saved, it's mainly used for tests
      SaveSourceCode(FileList()\FileName$, 0)
    EndIf
    
    If FileList()\UseMainFile
      FileNameToCompile$ = Filename$
      
      ; Little hack to simulate a file and get the flags from our main file
      ;
      AddElement(FileList())
      
      SourceFile = ReadFile(#PB_Any, FileNameToCompile$)
      If SourceFile
        FileLength = Lof()
        If FileLength > 4000
          FileLength = 4000
        EndIf
      
        FileSeek(Lof()-FileLength) ; the last 4000 bytes is more than enough for the project setting
        
        Buffer = AllocateMemory(FileLength+1)
        If Buffer
          ReadData(Buffer, FileLength)
          AnalyseProjectSettings(Buffer, FileLength)
          FreeMemory(Buffer)
        EndIf
      
        CloseFile(SourceFile)
      EndIf
      
      Flags = BuildCompilerFlags() & ~#PB_FLG_DEBUGGER  ; The debugger flags is dependant of the current file
      
      DeleteElement(FileList())
      
      If FileList()\Debugger
        Flags | #PB_FLG_DEBUGGER
      EndIf
    Else
      FileNameToCompile$ = FullPath + "Compilers\PB_EditorOutput.pb"
      SaveSourceCode(FileNameToCompile$, 1)
      Flags = BuildCompilerFlags()      
    EndIf
    
    If CreateFile(#CREATEFILE_COMPILERUN, SharedFileName)
      WriteStringN(FileNameToCompile$)
      WriteStringN(GetPathPart(FileName$))   ; Use the source code path, for executable default path
      WriteStringN(FileList()\CommandLine$)
      CloseFile(#CREATEFILE_COMPILERUN)
    EndIf
    
    PostThreadMessage_(CompilerThreadID, PB_MSG_ID, #PB_MSG_START_COMPILATION, Flags)
    
    IsProgramRunning = 1
  Else
    MessageRequester("Error", "The compiler isn't ready... Please try again.", #MB_ICONWARNING)
  EndIf
EndProcedure


Procedure Run()
  PostThreadMessage_(CompilerThreadID, PB_MSG_ID, #PB_MSG_RUN_PROGRAM, 0)
EndProcedure


Procedure CreateExecutable()
  
  If FileList()\ExecutableName$ = ""
    Path$ = SourceDirectory
  Else
    Path$ = FileList()\ExecutableName$
  EndIf
  
  CompilerIf #Demo
    If FileList()\ExecutableFormat$ = "Shared Dll"
      MessageRequester("Information", "DLL creation is not available in the demo version.", #MB_ICONINFORMATION)
    Else
  CompilerEndIf
  
  If FileList()\ExecutableFormat$ = "Shared Dll"
    Filter$ = "Shared DLL (*.dll)|*.DLL|All Files (*.*)|*.*"
    Extension$ = ".dll"    
  Else
    Filter$ = "Executable (*.exe)|*.EXE|All Files (*.*)|*.*"
    Extension$ = ".exe"
  EndIf
  
  If SaveSourceCode(FileList()\FileName$, 0)

    File$ = SaveFileRequester(Langage(156), Path$, Filter$, 0)
    If File$
      If LCase(Right(File$, 4)) <> Extension$
        File$+Extension$
      EndIf
      
      FileList()\ExecutableName$ = File$
      
      If CreateFile(#CREATEFILE_CREATEEXECUTABLE, SharedFileName)
        WriteStringN(File$)
        
        If FileList()\FileName$ = ""
          WriteStringN(FullPath+"Compilers\PB_EditorOutput.pb")
        Else
          WriteStringN(FileList()\FileName$)
        EndIf
        
        WriteStringN(GetPathPart(FileList()\FileName$))
        
        If FileList()\UseIcon
          WriteStringN(FileList()\IconName$)
        EndIf
        
        CloseFile(#CREATEFILE_CREATEEXECUTABLE)
        DisplayCompilingWindow()
        PostThreadMessage_(CompilerThreadID, PB_MSG_ID, #PB_MSG_CREATE_EXECUTABLE, BuildCompilerFlags())
        IsProgramRunning = 1;
      EndIf
    EndIf
  EndIf
      
  CompilerIf #Demo : EndIf : CompilerEndIf
    
EndProcedure


Procedure$ GetCurrentWord()
  
  SendMessage_(*RichEditGadget, #EM_EXGETSEL      , 1         , Range.CHARRANGE)
  NeededLine    = SendMessage_(*RichEditGadget, #EM_EXLINEFROMCHAR, 0         , Range\cpMin)
  StartPosition = SendMessage_(*RichEditGadget, #EM_LINEINDEX     , NeededLine, 0)
  
  *Buffer = AllocateMemory(1000)
  If *Buffer
    PokeW(*Buffer, 900) ; First word contains the max number of char for the buffer...
    
    LineLength       = SendMessage_(*RichEditGadget, #EM_GETLINE  , NeededLine, *Buffer)
    *CurrentLine.Byte = *Buffer+Range\cpMin-StartPosition
    
    If Range\cpMin > StartPosition
      
      Repeat
        *CurrentLine-1
        CurrentChar = *CurrentLine\b & $FF
      Until ValidCharacters(CurrentChar) = 0 Or Range\cpMin <= StartPosition
      
      *StartWord = *CurrentLine+1
    Else
      *StartWord = *CurrentLine
    EndIf
    
    Repeat
      *CurrentLine+1
      CurrentChar = *CurrentLine\b & $FF
    Until ValidCharacters(CurrentChar) = 0
    
    *CurrentLine\b = 0 ; Zero terminate the string, to allow the PeekS()
    
    Result$ = PeekS(*StartWord)
    
    FreeMemory(*Buffer)
  EndIf
  
  ProcedureReturn Result$
EndProcedure


Procedure DisplayHelp(HelpDirectory$, Topic$, HelpType)

  Select HelpType
    Case 1 ; PureBasic regular command
      If LCase(Right(HelpDirectory$, 4)) = ".chm"  ; A .chm has been defined in the .DESC -> It's a user lib with its own help
        Filename$ = FullPath+"Help\"+HelpDirectory$
        Topic$ +".html"
      Else
        Topic$    = CheckSpecialCommands(HelpDirectory$,Topic$)+".html"
        Filename$ = FullPath+"PureBasic.chm"
      EndIf
      OpenHelp(Filename$, Topic$)
      
    Case 2 ; Win32 API commands
      Filename$ = FullPath+"Help\Win32.hlp"
      If FileSize(Filename$) > 0
        OpenHelp(Filename$, Topic$)
      Else
        DisplayPlatformSDKHelp(Topic$)
      EndIf
      
    Default
      Filename$ = FullPath+"PureBasic.chm"
      Topic$    = CheckPureBasicKeyWords(Topic$)
      OpenHelp(Filename$, Topic$)
      
  EndSelect
    
EndProcedure


Procedure Help()
  
  PureFunctionHelp$ = GetCurrentWord()  ; Shared variable which will be used to receive the Help (in the event loop). Bad :).
  
  If FileList()\EnableASM And IsASMKeyword(PureFunctionHelp$)  ; Support for inline ASM sensitive help
    
    If Asc(UCase(PureFunctionHelp$)) = 'F'  ; Can only be an FPU related opcode
      OpenHelp(FullPath+"Help\ASMFPU.hlp", PureFunctionHelp$)
    Else
      OpenHelp(FullPath+"Help\ASM.hlp", PureFunctionHelp$)
    EndIf
    
  ElseIf CreateFile(#CREATEFILE_HELP, SharedFileName)
    WriteString(PureFunctionHelp$)
    CloseFile(#CREATEFILE_HELP)
    PostThreadMessage_(CompilerThreadID, PB_MSG_ID, #PB_MSG_Is_PureCommand, 0)
  EndIf
  
EndProcedure


Procedure UpdateProcedureList()
  TextLength = SendMessage_(*RichEditGadget, #WM_GETTEXTLENGTH, 0, 0)
  
  TextBuffer = AllocateMemory(TextLength+1)
  If TextBuffer
    SendMessage_(*RichEditGadget, #WM_GETTEXT, TextLength+1, TextBuffer)
    ProcedureParser(TextBuffer)
    FreeMemory(TextBuffer)
  EndIf
EndProcedure


Procedure UpdatePanelFilename()

  Filename$ = GetFilePart(FileList()\FileName$)

  If FileName$ = ""
    Filename$ = "<New>"
  EndIf

  If SendMessage_(*RichEditGadget, #EM_GETMODIFY, 0, 0) And Right(Filename$, 1) <> "*"
    Filename$+"*"
  EndIf
  
  If GetGadgetText(#GADGET_FilesPanel) <> Filename$  ; Change it only if needed (no flickering)
    SetGadgetText(#GADGET_FilesPanel, Filename$)
  EndIf
  
EndProcedure


Procedure UpdateToolTips()

  ToolBarToolTip(#MENU_New            , Langage(144))
  ToolBarToolTip(#MENU_Open           , Langage(145))
  ToolBarToolTip(#MENU_Save           , Langage(146))
  ToolBarToolTip(#MENU_Close          , Langage(147))
  ToolBarToolTip(#MENU_Cut            , Langage(148))
  ToolBarToolTip(#MENU_Copy           , Langage(149))
  ToolBarToolTip(#MENU_Paste          , Langage(150))
  ToolBarToolTip(#MENU_CompileRun     , Langage(151))
  ToolBarToolTip(#MENU_VisualDesigner , Langage(154))
  ToolBarToolTip(#MENU_Undo           , Langage(200))
  ToolBarToolTip(#MENU_Redo           , Langage(201))
  ToolBarToolTip(#MENU_Debugger       , Langage(202))
EndProcedure


Procedure QuickProcedureAccess()
  
  If EventType() = 7 ;And ProcedureBrowserType = 0
    UpdateProcedureList()
    
  ElseIf EventType() = 1 And ProcedureBrowserType = 0
    
    ItemIndex = SendMessage_(GadgetID(#GADGET_ProcedureList), #CB_GETCURSEL, 0, 0)
    If ItemIndex <> #CB_ERR
      ChangeActiveLine(SendMessage_(GadgetID(#GADGET_ProcedureList), #CB_GETITEMDATA, ItemIndex, 0), 0)
    EndIf
  ElseIf EventType() = #PB_EventType_LeftClick And ProcedureBrowserType = 1
    
    ItemIndex = SendMessage_(GadgetID(#GADGET_ProcedureBrowser), #LB_GETCURSEL, 0, 0)
    If ItemIndex <> #LB_ERR
      ChangeActiveLine(SendMessage_(GadgetID(#GADGET_ProcedureBrowser), #LB_GETITEMDATA, ItemIndex, 0), 0)
    EndIf
    SetFocus_(*RichEditGadget)
  EndIf
  
EndProcedure


Procedure SwitchSource(State)
  
  SendMessage_(*MainWindow, #WM_SETREDRAW, 0, 0)  ; Flickering fix
  
  SelectElement(FileList(), State)
  ShowWindow_(*RichEditGadget, #SW_HIDE)
  
  *RichEditGadget = FileList()\RichEdit

  ResizeRichEditGadget()
  
  UpdateDebuggerState()
  
  ShowWindow_(*RichEditGadget, #SW_SHOW)
  
  UpdateStatusBar2()

  If FileList()\FileName$
    SetWindowText_(*MainWindow, "PureBasic - "+FileList()\FileName$)
  Else
    SetWindowText_(*MainWindow, "PureBasic - <New>")
  EndIf

  OldLineNumberStart = -1 ; force update
  UpdateLineNumbering()

  SendMessage_(*MainWindow, #WM_SETREDRAW, 1, 0)  ; Flickering fix
  InvalidateRect_(*MainWindow, 0, 0)
  
  ProcedureBrowserMustUpdate = 1
  UpdateProcedureList()
  
  ; Check Selection in RichEdit window
  CheckForRichEditSelection()

  SetFocus_(*RichEditGadget)
EndProcedure

; Change the currently displayed file, useful for the keyboard shortcut (CTRL+TAB and CTRL+SHIFT+TAB)
;
Procedure SetCurrentFile(NewIndex)
  If NewIndex >= CountList(FileList())
    NewIndex = 0
  EndIf
  
  If NewIndex < 0
    NewIndex = CountList(FileList())-1
  EndIf
  
  If NewIndex <> ListIndex(FileList()) ; change only if necessary
    SwitchSource(NewIndex)
    SetGadgetState(#GADGET_FilesPanel, NewIndex)
  EndIf
EndProcedure
  

Procedure CloseCurrentFile(AlwaysOnePanel)
  
  If SourceCodeSaved()
    DestroyWindow_(FileList()\RichEdit)
    DeleteElement(FileList())
    RemoveGadgetItem(#GADGET_FilesPanel, GetGadgetState(#GADGET_FilesPanel))
    
    NbGadgetItems = CountGadgetItems(#GADGET_FilesPanel)
    
    If NbGadgetItems  = 0
      If AlwaysOnePanel
        CreateNewPanelItem("")
        SetFocus_(*RichEditGadget)
      EndIf
    Else
      LastElement(FileList())
      SwitchSource(NbGadgetItems-1)
      SetGadgetState(#GADGET_FilesPanel, NbGadgetItems-1)
    EndIf
    
    Result = 1
  EndIf
  
  ProcedureReturn Result
EndProcedure


Procedure CloseAllFiles()
  
  Result = 1
  
  While CountList(FileList()) And Result
    Result = CloseCurrentFile(0)
  Wend
  
  If (CountList(FileList()))
    SwitchSource(GetGadgetState(#GADGET_FilesPanel))
  EndIf
  
  ProcedureReturn Result
EndProcedure


Procedure CheckSavedBeforeExit()
  
  SavePreferences()
  
  If CloseAllFiles()
    Result = 1
  EndIf
  
  ProcedureReturn Result
EndProcedure


Procedure DisplayExternalError(Title$, Filename$, SkipHeader)

  If ReadFile(0, Filename$)

    If SkipHeader = #True
      ReadString() ; Skip the FASM Header
    EndIf
   
    While Eof(0) = 0
      ErrorDescription$ + ReadString() + Chr(10)
    Wend
    
    CloseFile(0)
  EndIf
  
  HideCompilingWindow()
  MessageRequester(Title$, ErrorDescription$, #MB_ICONERROR)
  
EndProcedure            


;- Editor Start

RichEditDLL = LoadLibrary_("RICHED20.DLL")

; Get the full program path, To make all the file access relative To it..
;

FullPath = Space(1000)
GetModuleFileName_(GetModuleHandle_(0), @FullPath, 1000)

FullPath = GetPathPart(FullPath)

SetCurrentDirectory_(FullPath) ; To fix the empty shortcut path stuff

SharedFileName.s = FullPath+"Compilers\Communication.msg"

LoadPreferences()
FileName$ = ProgramParameter()

RunOnce(FileName$)

; FullPath.s = "C:\Programmation\PureBasic\"

; Now, update the registry to associate correctly the '.pb' with the editor.
;

If GetVersion_() & $ff0000  ; Windows NT/XP
  
  If RegCreateKeyEx_(#HKEY_CLASSES_ROOT, "Applications\PureBasic.exe\shell\open\command", 0, 0, #REG_OPTION_NON_VOLATILE, #KEY_ALL_ACCESS, 0, @NewKey, @KeyInfo) = #ERROR_SUCCESS
    StringBuffer$ = Chr(34)+FullPath+"PureBasic.exe"+Chr(34)+" "+Chr(34)+"%1"+Chr(34)
    RegSetValueEx_(NewKey, "", 0, #REG_SZ,  StringBuffer$, Len(StringBuffer$)+1)
    RegCloseKey_(NewKey)
  EndIf
  
  If RegCreateKeyEx_(#HKEY_CURRENT_USER, "Software\Microsoft\Windows\CurrentVersion\Explorer\FileExts\.pb", 0, 0, #REG_OPTION_NON_VOLATILE, #KEY_ALL_ACCESS, 0, @NewKey, @KeyInfo) = #ERROR_SUCCESS
    RegSetValueEx_(NewKey, "Application", 0, #REG_SZ,  "PureBasic.exe", Len("PureBasic.exe")+1)
    RegCloseKey_(NewKey)
  EndIf
  
Else  ; The same for Win9x
  
  If RegCreateKeyEx_(#HKEY_LOCAL_MACHINE, "Software\Classes\PureBasic.exe\shell\open\command", 0, 0, #REG_OPTION_NON_VOLATILE, #KEY_ALL_ACCESS, 0, @NewKey, @KeyInfo) = #ERROR_SUCCESS
    StringBuffer$ = Chr(34)+FullPath+"PureBasic.exe"+Chr(34)+" "+Chr(34)+"%1"+Chr(34)
    RegSetValueEx_(NewKey, "", 0, #REG_SZ,  StringBuffer$, Len(StringBuffer$)+1)
    RegCloseKey_(NewKey)
  EndIf
  
  If RegCreateKeyEx_(#HKEY_LOCAL_MACHINE, "Software\CLASSES\.pb", 0, 0, #REG_OPTION_NON_VOLATILE, #KEY_ALL_ACCESS, 0, @NewKey, @KeyInfo) = #ERROR_SUCCESS
    RegSetValueEx_(NewKey, "", 0, #REG_SZ,  "PureBasic.exe", Len("PureBasic.exe")+1)
    RegCloseKey_(NewKey)
  EndIf
  
EndIf

; -------------------------------------------------------------

CurrentDirectory = SourceDirectory

Localize()

CreateGUI()

PB_MSG_ID = RegisterWindowMessage_("PB_MSG_ID")

Parameters.s = Str(GetCurrentThreadId_())+" /STANDBY"
CompilerProcessID = RunProgram(FullPath+"Compilers\PBCompiler.exe", Parameters, "", 2)
; CompilerProcessID = RunProgram("Compilers\PBCompiler.exe", Parameters, "", 2)

InitSyntaxHighlightning()

; Load a source code at startup If the command line is specified
; or if the AutoReload function is activated
;

NewSourceCode()

; Finally show the window, at the very end to avoid ugly flickering
;
;HideWindow(#WINDOW_Main,0)

If IsWindowMaximized
  ShowWindow_(*MainWindow, #SW_MAXIMIZE)
Else
  ShowWindow_(*MainWindow, #SW_SHOWNORMAL)
EndIf

CompilerIf #Demo = 1
  MessageRequester("Information", "This is the demo version of PureBasic "+#Version+"."+Chr(10)+"Please take a look to the 'Examples' drawer to test the programs"+Chr(10)+Chr(10)+"Demo limitation:"+Chr(10)+"- No Win32 API support"+Chr(10)+"- No DLL creation"+Chr(10)+"- Code size limitation (about 800 lines)"+Chr(10)+"- Debugger always enabled (much slower peformances, bigger executable)"+Chr(10)+Chr(10)+"Thanks a lot to take a look to PureBasic !"+Chr(10)+Chr(10)+"The Fantaisie Software Team.", #MB_ICONINFORMATION)
CompilerEndIf


If FileName$
  LoadSourceCodeReal(FileName$, 0)
  
  Repeat
    FileName$ = ProgramParameter()
    If Len(FileName$)
      LoadSourceCodeReal(FileName$, 0)
    EndIf
  Until FileName$ = ""
  
ElseIf AutoReload
  
  ForEach OpenedFiles()
    LoadSourceCodeReal(OpenedFiles(), 0)
  Next
  
EndIf

;- Kbd Shortcuts
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_F1 , #MENU_Help)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_F3 , #MENU_Next)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_F5 , #MENU_CompileRun)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_F12, #MENU_QuickLabelAccess)

AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_N | #PB_Shortcut_Control, #MENU_New)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_O | #PB_Shortcut_Control, #MENU_Open)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_S | #PB_Shortcut_Control, #MENU_Save)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_Z | #PB_Shortcut_Control, #MENU_Undo)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_Y | #PB_Shortcut_Control, #MENU_Redo)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_X | #PB_Shortcut_Control, #MENU_Cut)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_C | #PB_Shortcut_Control, #MENU_Copy)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_V | #PB_Shortcut_Control, #MENU_Paste)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_A | #PB_Shortcut_Control, #MENU_SelectAll)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_G | #PB_Shortcut_Control, #MENU_Goto)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_F | #PB_Shortcut_Control, #MENU_Find)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_R | #PB_Shortcut_Control, #MENU_Replace)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_W | #PB_Shortcut_Control, #MENU_Close)

; Shortcut to walk trough the opened file with the keyboard
;
AddKeyboardShortcut(#WINDOW_Main, #PB_ShortCut_Tab | #PB_Shortcut_Control, #MENU_NextOpenedFile)
AddKeyboardShortcut(#WINDOW_Main, #PB_ShortCut_Tab | #PB_Shortcut_Control | #PB_ShortCut_Shift, #MENU_PreviousOpenedFile)


; RichEdit comment/uncomment
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_B | #PB_Shortcut_Control , #MENU_CommentSelection  )
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_B | #PB_Shortcut_Alt     , #MENU_UnCommentSelection)

; Shortcuts for ToolMenu = ALT + Char
; Reserved Chars for Menu: F, E, C, T, H, V !!
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_V | #PB_Shortcut_Alt, #MENU_VisualDesigner)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_A | #PB_Shortcut_Alt, #MENU_ASCIITable)
AddKeyboardShortcut(#WINDOW_Main, #PB_Shortcut_S | #PB_Shortcut_Alt, #MENU_StructureViewer)
AddKeyboardShortcut(#WINDOW_Main, #PB_ShortCut_P | #PB_Shortcut_Alt, #MENU_PickColor)


#BufferLength = 1000
*Buffer = AllocateMemory(#BufferLength)

; Load the Cursor for the Splitter Bar (ProcedureBrowser)
; WinAPI: You cant free a cursor you got by LoadCursor_(),
; there isnt an API call for that.
Splitter_Cursor = LoadCursor_(0,#IDC_SIZEWE)

InvalidateRect_(*MainWindow, 0, 1)

;- Editor Events
Repeat
  EventID = WaitWindowEvent()
  
  Select EventWindowID()
  
    Case #WINDOW_About
      UseWindow( #WINDOW_About )
      AboutWindowEvents(EventID)
    
    Case #WINDOW_Goto
      UseWindow( #WINDOW_Goto )
      GotoWindowEvents(EventID)
    
    Case #WINDOW_Option
      UseWindow( #WINDOW_Option )
      OptionWindowEvents(EventID)
    
    Case #WINDOW_Find
      UseWindow( #WINDOW_Find )
      FindWindowEvents(EventID)
    
    Case #WINDOW_StructureViewer
      UseWindow( #WINDOW_StructureViewer )
      StructureViewerWindowEvents(EventID)
    
    Case #WINDOW_ASCIITable
      UseWindow( #WINDOW_ASCIITable )
      ASCIITableEvents(EventID)
    
    Case #WINDOW_PickColor
      UseWindow( #WINDOW_PickColor )
      PickColor_Events(EventID)
    
    Case #WINDOW_AddTools
      UseWindow(#WINDOW_AddTools)
      AddTools_Events(EventID)

    Case #WINDOW_Grep
      UseWindow(#WINDOW_Grep)
      GrepWindowEvents(EventID)

    Case #WINDOW_GrepOutput
      UseWindow(#WINDOW_GrepOutput)
      GrepOutputWindowEvents(EventID)
    
    Default
    
    If EventWindowID() = #WINDOW_Main
      UseWindow( #WINDOW_Main )
    EndIf
    
    Select EventID
        
        ;- Event PB_Message
      Case PB_MSG_ID
        
        Select EventwParam()
        
          Case #PB_MSG_Fatal_Error
            
            If ReadFile(#READFILE_PB_MSG_FATAL_ERROR, SharedFileName)
              ErrorDescription$ = ReadString()
              CloseFile(#READFILE_PB_MSG_FATAL_ERROR)
            EndIf
            
            MessageRequester("PureBasic - Compiler Error", ErrorDescription$, #MB_ICONERROR)
            End
            
            
          Case #PB_MSG_Assembler_Error
            DisplayExternalError("PureBasic - Assembler error", FullPath+"Compilers\Assembler.out", #True)
                     

          Case #PB_MSG_Linker_Error
            DisplayExternalError("PureBasic - Linker error", FullPath+"Compilers\Linker.out", #False)
            
            
          Case #PB_MSG_Resource_Error
            DisplayExternalError("PureBasic - Resource error", FullPath+"Compilers\Resource.out", #False)
                      
            
          Case #PB_MSG_Syntax_Error
            LineError = EventlParam()-1
            
            HideCompilingWindow()
            
            If ReadFile(#READFILE_PB_MSG_SYNTAX_ERROR, SharedFileName)
              ErrorDescription$ = ReadString()+Chr(10)+ReadString()  ; 2 lines are possible
              CloseFile(#READFILE_PB_MSG_SYNTAX_ERROR)
              
              If LineError > -1
                SendMessage_(*RichEditGadget, #EM_EXGETSEL, 1, Range.CHARRANGE)
                CurrentLineIndex = SendMessage_(*RichEditGadget, #EM_EXLINEFROMCHAR, 0, Range\cpMin)
                
                ChangeActiveLine(LineError+1, -3)
                LineErrorIndex = SendMessage_(*RichEditGadget, #EM_LINEINDEX, LineError, 0)
                SendMessage_(*RichEditGadget, #EM_SETSEL, LineErrorIndex, LineErrorIndex+SendMessage_(*RichEditGadget, #EM_LINELENGTH, LineErrorIndex, 0))
                
                ErrorDescription$ = Langage(25) + Str(LineError+1) + ": " + ErrorDescription$ ; 'Line XXX: Error XX'
              EndIf
              
              MessageRequester("PureBasic", ErrorDescription$, #MB_ICONERROR)
            EndIf
            
            SetFocus_(*RichEditGadget)
            IsProgramRunning = 0
            
            
          Case #PB_MSG_Compiler_Ready
            CompilerThreadID = EventlParam()
            PostThreadMessage_(CompilerThreadID, PB_MSG_ID, #PB_MSG_EDITOR_READY, 0)
            DisableToolBarButton(#MENU_CompileRun, 0)
            DisableMenuItem(#MENU_CompileRun, 0)
            DisableMenuItem(#MENU_StructureViewer, 0)
            LoadBasicFunctions()
            InitPureBasicFunctionsHT()
            
            
          Case #PB_MSG_Compilation_Finished
            HideCompilingWindow()
            IsProgramRunning = 0
            
            
          Case #PB_MSG_Is_PureCommand:
            
            If EventlParam()
              If ReadFile(#READFILE_PB_MSG_PURECOMMAND, SharedFileName)
                HelpDirectory$ = ReadString()  ; Return in which directory the help is (Window, Menu..) or a .chm file for user made libs
                
                CloseFile (#READFILE_PB_MSG_PURECOMMAND)
              EndIf
              
              If EventlParam() = 1 ; Build in command (including user one), not an API one.
                HelpPath$ = PureFunctionHelp$
              Else
                HelpPath$ = Left(PureFunctionHelp$, Len(PureFunctionHelp$)-1)  ; Remove the '_' at the end of the word
              EndIf
            EndIf
            
            DisplayHelp(HelpDirectory$, HelpPath$, EventlParam())
            
            
          Case #PB_MSG_Restart_Compiler ; external program trys to reload the compiler
            RestartCompiler()
             
             
            
        EndSelect
        
        ;- Event Gadgets
      Case #PB_EventGadget
        Select EventGadgetID()
            
          Case #GADGET_ProcedureList
            QuickProcedureAccess()
            
          Case #GADGET_ProcedureBrowser
            QuickProcedureAccess()
            
          Case #GADGET_FilesPanel
            SwitchSource(GetGadgetState(#GADGET_FilesPanel))
          
          Case #GADGET_ToolbarCloseButton
            CloseCurrentFile(1)
            
        EndSelect
        
        ;- Event popup menu
      Case #WM_RButtonDown
        If GetFocus_() = *RichEditGadget
          DisplayPopupMenu(1, WindowID(#WINDOW_Main))
        EndIf
        
        ;- Events splitter bar
      Case #WM_LBUTTONUP
        If Splitter_inMove
          ReleaseCapture_()
          Splitter_inMove = #FALSE
        EndIf
        
      Case #WM_LBUTTONDOWN
        x = EventlParam() & $FFFF
        y = EventlParam() >> 16
        GetClientRect_(*MainWindow, winrect.RECT)
        If ProcedureBrowserSide = 1 And x => EditLeftOffset - 6 And x <= EditLeftOffset And WindowMouseX() < EditLeftOffset*2
          Splitter_inMove = #TRUE
          SetCursor_( Splitter_Cursor )
          SetCapture_(*MainWindow)
        ElseIf ProcedureBrowserSide = 0 And x => winrect\right - EditRightOffset And x <= winrect\right - EditRightOffset + 4
          Splitter_inMove = #TRUE
          SetCursor_( Splitter_Cursor )
          SetCapture_(*MainWindow)
        Else
          Splitter_inMove = #FALSE
        EndIf
        
      Case #WM_MOUSEMOVE
        If EventwParam() = #MK_LBUTTON And Splitter_inMove
          x = EventlParam() & $FFFF
          y = EventlParam() >> 16
          SetCursor_( Splitter_Cursor )
          GetClientRect_(*MainWindow, winrect.RECT)
          If y > 28+EditTopOffset And y < winrect\bottom-20-29-EditBottomOffset
            If ProcedureBrowserSide = 1 ; Left side
              If x > 20 And x < winrect\right - 20
                EditLeftOffset = x
                ProcedureBrowserWidth = EditLeftOffset
                SendMessage_(*MainWindow, #WM_SIZE, #SIZE_RESTORED,0)
              EndIf
            Else
              If x > 20 And x < winrect\right - 20
                EditRightOffset = winrect\right - ( EventlParam() & $FFFF )
                ProcedureBrowserWidth = EditRightOffset
                SendMessage_(*MainWindow, #WM_SIZE, #SIZE_RESTORED,0)
              EndIf
            EndIf
          EndIf
        Else
          If ProcedureBrowserType
            x = EventlParam() & $FFFF
            y = EventlParam() >> 16
            GetClientRect_(*MainWindow, winrect.RECT)
            If y > 28+EditTopOffset And y < winrect\bottom-20-29-EditBottomOffset
              If ProcedureBrowserSide = 1 And x => EditLeftOffset - 6 And x <= EditLeftOffset And WindowMouseX() < EditLeftOffset*2
                SetCursor_( Splitter_Cursor )
              ElseIf ProcedureBrowserSide = 0 And x => winrect\right - EditRightOffset And x <= winrect\right - EditRightOffset + 4
                SetCursor_( Splitter_Cursor )
              Else
                Splitter_inMove = 0
              EndIf
            Else
              Splitter_inMove = 0
            EndIf
          EndIf
        EndIf
        
        ;- Event DropFiles
      Case #WM_DropFiles
        
        *DropHandle = EventwParam()
        If *DropHandle
          NbDroppedFiles = DragQueryFile_ (*DropHandle, $FFFFFFFF, *Buffer, #BufferLength)
          
          For k = 0 To NbDroppedFiles-1
            DragQueryFile_(*DropHandle, k, *Buffer, #BufferLength)
            LoadSourceCodeReal(PeekS(*Buffer), 0)
          Next
          
          DragFinish_(*DropHandle)
        EndIf
        
        
        ;- Event Menus
      Case #PB_EventMenu
        MenuID = EventMenuID()
        Select MenuID
            
          Case #MENU_New
            NewSourceCode()
            
          Case #MENU_Open
            LoadSourceCode()
            
          Case #MENU_Save
            SaveSourceCode(FileList()\FileName$, 0)
            SetFocus_(*RichEditGadget) ; To re-activate the RichEdit after the save as dialog, in the case of a new source file.
            
          Case #MENU_SaveAs
            SaveSourceCodeAs()
            SetFocus_(*RichEditGadget) ; To re-activate the RichEdit after the save as dialog.
            
          Case #MENU_Close
            CloseCurrentFile(1)
            
          Case #MENU_Preference
            OpenPreferenceWindow()
            
          Case #MENU_Undo
            Undo()
            
          Case #MENU_Redo
            Redo()
            
          Case #MENU_Cut
            ClipboardCut()
            
          Case #MENU_Copy
            ClipboardCopy()
            
          Case #MENU_Paste
            ClipboardPaste()
            
          Case #MENU_CommentSelection
            CommentSelection()
            
          Case #MENU_UnCommentSelection
            UnCommentSelection()
            
          Case #MENU_SelectAll
            SelectAll_Range.CHARRANGE\cpMin = 0
            SelectAll_Range\cpMax = -1
            SendMessage_(*RichEditGadget, #EM_EXSETSEL,0,@SelectAll_Range)
            
          Case #MENU_Goto
            OpenGotoWindow()
            
          Case #MENU_Find
            OpenFindWindow(0)
            
          Case #MENU_Replace
            OpenFindWindow(1)
            
          Case #MENU_Next
            FindTextNext(3)
            SetFocus_(*RichEditGadget) ; To re-activate the RichEdit if the text isn't found (after the MessageRequester()).

          Case #MENU_FindInFiles
            OpenGrepWindow()
            
          Case #MENU_CompileRun
            CompileRun()
            
          Case #MENU_Run
            Run()
            
          Case #MENU_Debugger
            FileList()\Debugger = 1-GetMenuItemState(0, #MENU_Debugger)
            UpdateDebuggerState()
            
          Case #MENU_CompilerOption:
            OpenOptionWindow()
            
          Case #MENU_CreateExecutable:
            CreateExecutable()
            
          Case #MENU_RestartCompiler
            RestartCompiler()
            
          Case #MENU_StructureViewer:
            If CompilerThreadID           ; The compiler need to be loaded before called this.
              OpenStructureViewerWindow()
            EndIf
            
          Case #MENU_VisualDesigner
            RunProgram(VisualDesigner$)
            
          Case #MENU_ASCIITable
            OpenASCIITable()
            
          Case #MENU_PickColor
            PickColor_Open()
            
          Case #MENU_SmartUpdate
            If FileSize(FullPath+"SmartUpdate.exe") > 0
              If Editor_RunOnce <> 0 Or MessageRequester("PureBasic", Langage(170), #MB_OKCANCEL) = #IDOK  
                Quit = CheckSavedBeforeExit() ; Editor has to quit, so he can be updated by the tool.
                If Quit
                  Quit = RunProgram("SmartUpdate.exe", "/EDITOR", FullPath) ; If the RunProgram fails, don't quit !
                EndIf
              EndIf
            Else
              MessageRequester("PureBasic", Langage(198), #MB_ICONERROR) ; "SmartUpdater tool not found in the PureBasic directory."
            EndIf
            
          Case #MENU_AddTools
            AddTools_Open()
            
          Case #MENU_Help
            Help()
            
          Case #MENU_About:
            OpenAboutWindow()
            
          Case #MENU_Exit:
            Quit = CheckSavedBeforeExit()
            
          Case #MENU_QuickLabelAccess
            If ProcedureBrowserType = 0
              SendMessage_(GadgetID(#GADGET_ProcedureList), #CB_SHOWDROPDOWN, 1, 0)
              ActivateGadget(#GADGET_ProcedureList)
              QuickProcedureAccess()
            Else
              ProcedureBrowserMustUpdate = 1
              UpdateProcedureList()
              ;ActivateGadget(#GADGET_ProcedureBrowser)
            EndIf
            
          Case #MENU_NextOpenedFile
            SetCurrentFile(ListIndex(FileList())+1)

          Case #MENU_PreviousOpenedFile
            SetCurrentFile(ListIndex(FileList())-1)
            
          Default
            
            ; Additional Help Files are Menu 100 - 200
            If MenuID => #MENU_HelpFilesStart And MenuID <= #MENU_HelpFilesEnd
              Menu_OpenHelpFile(MenuID)
            EndIf
            
            ; External Addon Tools
            If MenuID => #MENU_AddToolsStart And MenuID <= #MENU_AddToolsEnd
              AddTools_MenuClicked(MenuID)
            EndIf
            
            If MenuID => #MENU_RecentFilesStart And MenuID <= #MENU_RecentFilesEnd
              Menu_OpenRecentFile(MenuID)
            EndIf
            
        EndSelect
        
      Case #PB_EventCloseWindow
        Quit = CheckSavedBeforeExit()
        
    EndSelect
    
  EndSelect
  
Until Quit

;-Editor End

; Mutex for RunOnce (Run one instance only)
If Editor_RunOnceMutex
  CloseHandle_(Editor_RunOnceMutex)
EndIf

TerminateProcess_(CompilerProcessID,0)  ; Terminate the PB Compiler process
QuitEditor()
DeleteFile(SharedFileName)

If *LineNumbers
  DestroyWindow_(*LineNumbers)
EndIf
; Free RichEdit DLL (needed for Win95 to not crash at end)
FreeLibrary_(RichEditDLL)
; Delete Brush for ProcedureBrowser ListView
DeleteObject_(ProcedureBrowserBackBrush)
End
; ExecutableFormat=Windows
; EnableXP
; UseIcon=Data\PureBasic IDE.ico
; Executable=C:\Programmation\PureBasic\PureBasic2.exe
; EOF