;
; < ------------------------------------ >
;   PureBasic IDE - Syntax Highlightning 
; < ------------------------------------ >
;
; This program is free software; you can redistribute it and/or modify it under
; the terms of the GNU General Public License as published by the Free Software
; Foundation; either version 2 of the License, or (at your option) any later
; version. See the License.txt file for more details.
; 
; (c) 2003 Fantaisie Software
;
XIncludeFile "Common.pb"
XIncludeFile "Misc.pb"

Global BasicKeyword$, ASMKeyword$  ; WARNING ! Can be very dangerous

#NbASMKeywords = 232

#tomSuspend = -9999995 ; richedit 3.0 text object model constants
#tomResume  = -9999994

Declare BuildRTFBuffer(*TextBuffer.Byte, TextBufferLength)
Declare StreamCallBack(dwCookie, pbBuff, cb, *pcb)

Dim BasicKeywordsHT.l(255)
Dim BasicKeywords.s(#NbBasicKeywords)
Dim BasicKeywordsReal.s(#NbBasicKeywords)

Dim ASMKeywords.s(#NbASMKeywords)
Dim ASMKeywordsHT.l(255)

Dim BasicFunctionsHT.l(255)
Dim APIFunctionsHT.l(255)
Dim APIFunctionsHT2.l(255)    ; contains the last index for each char


Procedure InitSyntaxHighlightning()

  ; Build the ValidCharacter table to have have a faster routine to detect
  ; if the char is a considered as a string or not. Nice improvement :)
 
  For k='0' To '9'            ; Tell than the ASCII from 0 to 9 are valid
    ValidCharacters(k) = 1
  Next

  For k='A' To 'Z'            ; Tell than the ASCII from A to Z are valid
    ValidCharacters(k) = 1
  Next

  For k='a' To 'z'            ; Tell than the ASCII from a to z are valid
    ValidCharacters(k) = 1
  Next

  ValidCharacters('_') = 1

  ; Now, init the internal PureBasic keywords

  CurrentChar = 0
  Restore BasicKeywords
  For k=1 To #NbBasicKeywords
    Read BasicKeywordsReal(k)
    Read BasicKeywords(k)
    
    Char = Asc(BasicKeywords(k))
    If Char <> CurrentChar
      BasicKeywordsHT(Char) = k
      CurrentChar = Char
    EndIf
  Next

  ; And next, init the ASM keywords

  CurrentChar = 0
  Restore ASMKeywords
  For k=1 To #NbASMKeywords
    Read ASMKeywords(k)
    
    Char = Asc(ASMKeywords(k))
    If Char <> CurrentChar
      ASMKeywordsHT(Char) = k
      CurrentChar = Char
    EndIf
  Next
  
  ; Finally, init the API functions
  
  LoadAPIFunctions()

  ; API function Hash Table

  CurrentChar = 0
  For k=1 To NbAPIFunctions
    Char = Asc(UCase(PeekS(*APIFunctionsArray(k-1,0))))
    If Char <> CurrentChar
      APIFunctionsHT(Char) = k
      APIFunctionsHT2(CurrentChar) = k-2
      CurrentChar = Char
    EndIf
  Next
  
EndProcedure


Procedure InitPureBasicFunctionsHT()

  ; PureBasic function Hash Table

  CurrentChar = 0
  For k=1 To NbBasicFunctions
    ;Debug *BasicFunctionsArray(k,0)
    ;Debug PeekS(*BasicFunctionsArray(k-1,0))
    Char = Asc(UCase(PeekS(*BasicFunctionsArray(k-1,0))))
    If Char <> CurrentChar
      BasicFunctionsHT(Char) = k
      CurrentChar = Char
    EndIf
  Next

EndProcedure


Procedure IsBasicKeyword(Word$)

  Word$ = LCase(Word$)
  k = BasicKeywordsHT(Asc(Word$))  ; We use an HashTable to get access to the index very quickly
  If k
    While Quit = 0 And k <= #NbBasicKeywords
      
      If BasicKeywords(k) <= Word$
        If BasicKeywords(k) = Word$
          BasicKeyword$ = BasicKeywordsReal(k)
          Result = k
        EndIf
      Else
        Quit = 1
      EndIf

      k+1
    Wend
  EndIf

  ProcedureReturn Result
EndProcedure


Procedure IsBasicKeyword_ProgramFlow(Word$)
  If Word$ = "If" Or Word$ = "Else" Or Word$ = "ElseIf" Or Word$ = "EndIf"
    result = 1
  ElseIf Word$ = "While" Or Word$ = "Wend" Or Word$ = "For" Or Word$ = "To" Or Word$ = "Step" Or Word$ = "Next"
    result = 1
  ElseIf Word$ = "Break" Or Word$ = "Continue" Or Word$ = "ForEach"
    result = 1
  ElseIf Word$ = "Repeat" Or Word$ = "Until" Or Word$ = "Forever"
    result = 1
  ElseIf Word$ = "Select" Or Word$ = "Case" Or Word$ = "Default" Or Word$ = "EndSelect"
    result = 1
  ; dunno about this:
  ElseIf Word$ = "Goto" Or Word$ = "Gosub" Or Word$ = "Or" Or Word$ = "And"
    result = 1
  EndIf
  ProcedureReturn result
EndProcedure


Procedure IsASMKeyword(Word$)

  Word$ = UCase(Word$)
  k = ASMKeywordsHT(Asc(Word$))  ; We use an HashTable to get access to the index very quickly
  If k
    While Quit = 0 And k <= #NbASMKeywords
      
      If ASMKeywords(k) <= Word$
        If ASMKeywords(k) = Word$
          ASMKeyword$ = ASMKeywords(k)
          Result = k
        EndIf
      Else
        Quit = 1
      EndIf

      k+1
    Wend
  EndIf

  ProcedureReturn Result
EndProcedure


Procedure IsBasicFunction(Word$)
  Result = -1
  
  k = BasicFunctionsHT(Asc(UCase(Word$)))  ; We use an HashTable to get access to the index very quickly
  If k
    k-1
    While Quit = 0 And k < NbBasicFunctions
    
      Compare = CompareMemoryString(*BasicFunctionsArray(k,0), Word$, 1)  ; Case insensitive compare
    
      If Compare <= 0
        If Compare = 0
          Result = k
          Quit   = 1
        EndIf
      Else
        Quit = 1
      EndIf

      k+1
    Wend
  EndIf
  
  ProcedureReturn Result
EndProcedure



Procedure FindAPIFunctionFast(First, Last, *Word)

  If Last = First+1
    If CompareMemoryString(*Word, *APIFunctionsArray(First, 0), 1)
      ProcedureReturn First
      
    ElseIf CompareMemoryString(*Word, *APIFunctionsArray(Last, 0), 1)
      ProcedureReturn Last
      
    Else
      ProcedureReturn -1
     
    EndIf
  EndIf

  Index = (First+Last)/2
  Result = CompareMemoryString(*Word, *APIFunctionsArray(Index, 0), 1)
  
  If Result = 0
    ProcedureReturn Index
    
  ElseIf Result = 1
    ProcedureReturn FindAPIFunctionFast(Index, Last, *Word)
    
  ElseIf Result = -1
    ProcedureReturn FindAPIFunctionFast(First, Index, *Word)
  
  EndIf

EndProcedure


; 
; Procedure IsAPIFunction(Word$)
;   Result = -1
;   
;   If Right(Word$, 1) = "_"
;     Word$ = Left(Word$, Len(Word$)-1)
;     
;     k = APIFunctionsHT(Asc(UCase(Word$)))  ; We use an HashTable to get access to the index very quickly
;     If k
;       ; because there are many API functions, a faster routine is needed:
;       Result = FindAPIFunctionFast(k-1, APIFunctionsHT2(Asc(UCase(Word$))), @Word$)      
;     EndIf
;     
;   EndIf
;   
;   ProcedureReturn Result
; EndProcedure

Procedure IsAPIFunction(Word$)
  Result = -1
  
  If Right(Word$, 1) = "_"
    Word$ = Left(Word$, Len(Word$)-1)
  
    k = APIFunctionsHT(Asc(UCase(Word$)))  ; We use an HashTable to get access to the index very quickly
    If k
      k-1
      While Quit = 0 And k < NbAPIFunctions
      
        Compare = CompareMemoryString(*APIFunctionsArray(k,0), Word$, 1)  ; Case insensitive compare
      
        If Compare <= 0
          If Compare = 0
            Result = k
            Quit   = 1
          EndIf
        Else
          Quit = 1
        EndIf

        k+1
      Wend
    EndIf
  
  EndIf
    
  ProcedureReturn Result
EndProcedure




Procedure IsDecNumber(SearchString$)
  *string.BYTE = @SearchString$
  While (*string\b & $FF) <> 0
    If (*string\b & $FF) < 48 Or (*string\b & $FF) > 57
      ProcedureReturn 0
    EndIf
    *string + 1
  Wend
  ProcedureReturn 1
EndProcedure


Procedure ClearEffects()

  CharacterFormat.CHARFORMAT
  CharacterFormat\cbSize      = SizeOf(CHARFORMAT)
  CharacterFormat\dwMask      = #CFM_COLOR | #CFM_BOLD
  CharacterFormat\crTextColor = NormalTextColor

  SendMessage_(*RichEditGadget, #EM_SETCHARFORMAT, #SCF_SELECTION, CharacterFormat)
  
EndProcedure


Procedure SyntaxHighlighting()

  If EnableColoring
  
    ; Suspend the undo buffer during the hilighting - requires richedit 3.0
    ; other versions simply don't know the message
    ;
    SendMessage_(*RichEditGadget, #EM_GETOLEINTERFACE, 0, @RichEditOleObject.IRichEditOle)
    If RichEditOleObject
      If RichEditOleObject\QueryInterface(?IID_ITextDocument, @TextInterface.ITextDocument) <> #E_NOINTERFACE
        TextInterface\Undo(#tomSuspend, 0) 
      EndIf
    EndIf
    
    HorizontalScrollBarPosition = GetScrollPos_(*RichEditGadget, #SB_HORZ)

      
    SendMessage_(*RichEditGadget, #WM_SETREDRAW, 0, 0)
    SendMessage_(*RichEditGadget, #EM_SETOPTIONS, #ECOOP_XOR, #ECO_AUTOHSCROLL)  ; Remove the autoscroll feature, as if the line is too long, the SETSEL will move back the scrollbar at the beginning of the line
      
                 SendMessage_ (*RichEditGadget, #EM_HIDESELECTION , 1, 0)
                 SendMessage_ (*RichEditGadget, #EM_EXGETSEL      , 1, Range.CHARRANGE)
    NeededLine = SendMessage_ (*RichEditGadget, #EM_EXLINEFROMCHAR, 1, Range\cpMin)
    
    If Range\cpMin = Range\cpMax

      *BufferHeader = *LineBuffer  ; Fix the amount of characters which can be read..
      PokeL(*BufferHeader, 1000)

      StartPosition = SendMessage_ (*RichEditGadget, #EM_LINEINDEX, NeededLine, 0);
      LineLength    = SendMessage_ (*RichEditGadget, #EM_GETLINE  , NeededLine, *LineBuffer)
      
      
      XYPosition = SendMessage_(*RichEditGadget, #EM_POSFROMCHAR, StartPosition, 0)
            
      *Cursor.Byte = *LineBuffer
      LineLength = 0
      While *Cursor\b <> 10 And *Cursor\b <> 0 And *Cursor\b <> 13
        *Cursor+1
        LineLength+1
      Wend
      
      *CurrentLine = *LineBuffer
      PokeB(*LineBuffer+LineLength, 0)

      If LineLength > 0
          
        *RTFBuffer = BuildRTFBuffer(*CurrentLine, LineLength)
        SendMessage_(*RichEditGadget, #EM_SETSEL, StartPosition, StartPosition+LineLength);  ; MODIFIED here for v20.dll should be: StartPosition+LineLength+1)

        *CallbackBuffer       = *RTFBuffer
         CallbackBufferLength = MemoryStringLength(*CallbackBuffer)

        StreamStructure.EDITSTREAM\pfnCallback = @StreamCallBack()
        
        SendMessage_(*RichEditGadget, #EM_STREAMIN, #SF_RTF | #SFF_SELECTION, StreamStructure)
        GlobalFree_(*RTFBuffer)
      EndIf
    EndIf
    
    SendMessage_(*RichEditGadget, #EM_SETSEL, Range\cpMin, Range\cpMin) ; Put the pointer to it's right place...
    SendMessage_(*RichEditGadget, #EM_HIDESELECTION, 0, 0)
    SendMessage_(*RichEditGadget, #EM_SETOPTIONS, #ECOOP_OR, #ECO_AUTOHSCROLL)
    SetFocus_   (*RichEditGadget)
    
    SendMessage_(*RichEditGadget, #WM_SETREDRAW, 1, 0)
    
    ; It's a line by line SyntaxHightlightning, so refresh only the part which need it !
    ; It avoid slight flickering when typing
    ;
    ;If HorizontalScrollBarPosition = 0 ; Works only if we are at the start of the line, else the whole screen can be redrawed if we do an up/down
    ;  ModifiedRect.RECT
    ;  ModifiedRect\Top    = XYPosition >> 16 - EditorFontSize*2
    ;  ModifiedRect\Right  = 10000                ; 10000 pixels max width for RichEdit.. :)
    ;  ModifiedRect\Bottom = XYPosition >> 16 + EditorFontSize*2 ; Little hack, the EditorFontSize seems to not be exactly the height !
    ;  InvalidateRect_(*RichEditGadget, ModifiedRect, 1)
    ;Else
      InvalidateRect_(*RichEditGadget, 0, 1)
    ;EndIf
       
    ; resume the undo buffer's operation and release the object
    If TextInterface
      TextInterface\Undo(#tomResume, 0)
      TextInterface\Release()
    EndIf
    If RichEditOleObject
      RichEditOleObject\Release()
    EndIf    
    
  EndIf
EndProcedure


Procedure UpdateTextHighlightning(IsEnable, *TextBuffer, TextLength)
  Shared *RichEditGadget
  
  ; Suspend the undo buffer. requires richedit 3.0
  ; other versions simply don't know the message
  ;
  SendMessage_(*RichEditGadget, #EM_GETOLEINTERFACE, 0, @RichEditOleObject.IRichEditOle)
  If RichEditOleObject
    If RichEditOleObject\QueryInterface(?IID_ITextDocument, @TextInterface.ITextDocument) <> #E_NOINTERFACE
      TextInterface\Undo(#tomSuspend, 0) 
    EndIf
  EndIf    

  If *TextBuffer = 0  ; This must be here, as even unhighlightned text can be copied when loading..
 
    TextLength = SendMessage_(*RichEditGadget, #WM_GETTEXTLENGTH, 0, 0)
   *TextBuffer = LocalAlloc_ (#LMEM_ZEROINIT , TextLength+1)
                 SendMessage_(*RichEditGadget, #WM_GETTEXT, TextLength+1, *TextBuffer)
    Allocated = 1
  EndIf

  SendMessage_(*RichEditGadget, #EM_SETBKGNDCOLOR, 0, BackgroundColor);

  If IsEnable
  
    *RTFBuffer = BuildRTFBuffer(*TextBuffer, TextLength)
    If *RTFBuffer
      *CallbackBuffer = *RTFBuffer
      CallbackBufferLength = MemoryStringLength(*CallbackBuffer)
      
      StreamStructure.EDITSTREAM\pfnCallback = @StreamCallBack()
      SendMessage_(*RichEditGadget, #EM_STREAMIN, #SF_RTF, StreamStructure)
      
      GlobalFree_(*RTFBuffer)
    EndIf
  Else
    ClearEffects()
    SendMessage_(*RichEditGadget, #WM_SETTEXT, 0, *TextBuffer)
  EndIf

  If Allocated
    LocalFree_(*TextBuffer)
  EndIf
  
  ; resume the undo buffer's operation and release the object
  If TextInterface
    TextInterface\Undo(#tomResume, 0)
    TextInterface\Release()
  EndIf
  If RichEditOleObject
    RichEditOleObject\Release()
  EndIf 
  
EndProcedure


Procedure AddRTFTag(Tag$)
  CopyMemoryString(Tag$)
EndProcedure

Procedure BuildRTFColor(Color)

  CopyMemoryString("\red"+Str(Color & $FF))
  CopyMemoryString("\green"+Str((Color >> 8) & $FF))
  CopyMemoryString("\blue"+Str((Color >> 16) & $FF))
  CopyMemoryString(";")

EndProcedure


Procedure BuildRTFBuffer(*TextBuffer.Byte, TextBufferLength)

  RTFBufferSize = 200000  ; enough for most sources

 *RTFBuffer.Byte = GlobalAlloc_(#GMEM_FIXED, RTFBufferSize);
 *RTFBufferStart = *RTFBuffer;
 
  acp.l = GetACP_()
  If acp=0 : acp=1252 : EndIf ; Default if none available?
  ;MessageRequester("CodePage", Str(acp), #PB_MessageRequester_OK)
  CopyMemoryString("{\rtf1\ansi\ansicpg"+Str(acp)+"\uc1 \deff0{\colortbl;", @*RTFBuffer) ; This is the first call to CopyMemoryString(), so we pass the start of the memory buffer
                                                          
  BuildRTFColor(NormalTextColor)
  BuildRTFColor(BasicKeywordColor)
  BuildRTFColor(CommentColor)
  BuildRTFColor(ConstantColor)
  BuildRTFColor(StringColor)
  BuildRTFColor(PureKeywordColor)
  BuildRTFColor(ASMKeywordColor)
  BuildRTFColor(OperatorColor)
  BuildRTFColor(StructureColor)
  BuildRTFColor(NumberColor)
  BuildRTFColor(PointerColor)
  BuildRTFColor(SeparatorColor)
  BuildRTFColor(LabelColor)
  BuildRTFColor(StringColor2) ; A$
  BuildRTFColor(BasicKeywordColor_ProgramFlow) ; If..Else..EndIf, While..Wend, ...

  If EditorFontStyle & #PB_FONT_BOLD
    CopyMemoryString("}\b\cf1 ")
  Else
    CopyMemoryString("}\b0\cf1 ")
  EndIf


  NewLine = 0
  *LineStart = *TextBuffer
  
  EnableASM = FileList()\EnableASM
  
  LineStart = 1      

  While *TextBuffer\b <> 0
  
    ; Skip all the spaces before the word And fill the *RTFBuffer with spaces..
    
    While *TextBuffer\b = ' ' Or *TextBuffer\b = 9  ; Space or Tab
      *RTFBuffer\b = ' '
      *RTFBuffer+1
      *TextBuffer+1
    Wend
    
    ; Now, isolate the word..

    *WordStart = *TextBuffer

    While(ValidCharacters(*TextBuffer\b & $FF))
      *TextBuffer+1
    Wend

    *WordEnd.Byte = *TextBuffer;

    ; Skip all the spaces after the word..

    AfterSpaces = 0;
    While *TextBuffer\b = 32 Or *TextBuffer\b = 9
      AfterSpaces+1
      *TextBuffer+1
    Wend

    ; Here we got our separator: (, ', ;....). If it's a valid character, then the separator was spaces..

    If ValidCharacters(*TextBuffer\b & $FF) Or *TextBuffer\b = 13 Or *TextBuffer\b = 10 Or *TextBuffer\b = 0
      If *TextBuffer\b = 13 Or *TextBuffer\b = 10
        NewLine = 1
      EndIf

      If *TextBuffer\b = 13
        *TextBuffer+1
      EndIf

      SeparatorChar = 1
    Else
      SeparatorChar = *TextBuffer\b
    EndIf

    *WordEnd\b = 0

    WordStart$ = PeekS(*WordStart)
    
    Flagged  = 0
    IsString = 0
    UseSpecialOperator = 0

    If WordStart$
      If SeparatorChar = '('
      
        ; check for API first, because this check is fast if it is not an API function (quits imediately because of the missing '_')
        FunctionPosition = IsAPIFunction(WordStart$)
        
        If FunctionPosition > -1
          If CursorPositionX >= *WordStart-*LineStart And CursorPositionX <= *TextBuffer-*LineStart+2
            QuickHelpPosition = FunctionPosition
            IsQuickHelpWinAPI = 1
          EndIf

          PokeS(WordStart$, PeekS(*APIFunctionsArray(FunctionPosition, 0))+"_") ; add the '_', as it is not presend in the funcions array
          
        Else 
        
          FunctionPosition = IsBasicFunction(WordStart$)
          
          If FunctionPosition > -1
            If CursorPositionX >= *WordStart-*LineStart And CursorPositionX <= *TextBuffer-*LineStart+2
              QuickHelpPosition = FunctionPosition
              IsQuickHelpWinAPI = 0
            EndIf

            PokeS(WordStart$, PeekS(*BasicFunctionsArray(FunctionPosition, 0))) ; Faster than WordStart$ = PeekS(..), because no reallocation
          EndIf        
        
        EndIf        
        

        If WordStart$ <> "b" And WordStart$ <> "w" And WordStart$ <> "l" And WordStart$ <> "f" And WordStart$ <> "s" And WordStart$ <> "B" And WordStart$ <> "W" And WordStart$ <> "L" And WordStart$ <> "F" And WordStart$ <> "S"
          If OldSeparatorChar = '.' ; example: Dim BasicKeywordsHT.l(255)
            AddRTFTag("\cf9 ")      ; DIM x.Structure(100)
          Else
            AddRTFTag("\cf6 ")      ; PB Functions
          EndIf
          Flagged = 1
        EndIf
      ; Structure Access
      ElseIf SeparatorChar = '\' Or OldSeparatorChar = '\' Or (OldSeparatorChar = '.' And PB_TYPE_DETECTED = #FALSE)
        
        If SeparatorChar = '$' And AfterSpaces = 0 And IsBasicKeyword(WordStart$)=0
          *TextBuffer + 1
          WordStart$ + "$"
          AddRTFTag("\cf14 ")
          SeparatorChar      = 1
          IsString           = 1
        ElseIf IsDecNumber(WordStart$)
          AddRTFTag("\cf10 ")
        Else
          A$ = UCase(WordStart$)
          If A$ <> "B" And A$ <> "W" And A$ <> "L" And A$ <> "F" And A$ <> "S"
            AddRTFTag("\cf9 ")
          EndIf
        EndIf
        UseSpecialOperator = 1
        Flagged            = 1
      ; Structure Declaration
      ElseIf SeparatorChar = '.'
        PB_TYPE_DETECTED = #FALSE
        If IsDecNumber(WordStart$) ; Float -> 1.2
          AddRTFTag("\cf10 ")
          Flagged            = 1
          UseSpecialOperator = 1
        Else
          *AfterSeparator.Byte = *TextBuffer + 1
          If ValidCharacters(*AfterSeparator\b & $FF)
            If (*AfterSeparator\b & $FF) = 'b' Or (*AfterSeparator\b & $FF) = 'w' Or (*AfterSeparator\b & $FF) = 'l' Or (*AfterSeparator\b & $FF) = 'f' Or (*AfterSeparator\b & $FF) = 's' Or (*AfterSeparator\b & $FF) = 'B' Or (*AfterSeparator\b & $FF) = 'W' Or (*AfterSeparator\b & $FF) = 'L' Or (*AfterSeparator\b & $FF) = 'F' Or (*AfterSeparator\b & $FF) = 'S'
              *AfterSeparator + 1
              While (*AfterSeparator\b & $FF) = ' ' : *AfterSeparator + 1 : Wend
              If ValidCharacters(*AfterSeparator\b & $FF)
                While ValidCharacters(*AfterSeparator\b & $FF) Or (*AfterSeparator\b & $FF) = ' '
                  *AfterSeparator + 1
                Wend
                If (*AfterSeparator\b & $FF) = '(' ; DIM xyz.point(
                  AddRTFTag("\cf6 ")               ; NewList xyz.point(
                  Flagged = 1
                Else
                  AddRTFTag("\cf9 ")               ; Structure
                  Flagged = 1
                EndIf
              ElseIf (*AfterSeparator\b & $FF) = '('
                AddRTFTag("\cf6 ")
                Flagged = 1
              Else
                PB_TYPE_DETECTED = #TRUE
              EndIf
            Else
              While ValidCharacters(*AfterSeparator\b & $FF) Or (*AfterSeparator\b & $FF) = ' '
                *AfterSeparator + 1
              Wend
              If (*AfterSeparator\b & $FF) = '(' ; DIM xyz.point(
                AddRTFTag("\cf6 ")               ; NewList xyz.point(
                Flagged = 1
              Else
                AddRTFTag("\cf9 ")               ; Structure
                Flagged = 1
              EndIf
            EndIf
          EndIf
        EndIf
      ; String$
      ElseIf SeparatorChar = '$' And AfterSpaces = 0 And IsBasicKeyword(WordStart$)=0
        *TextBuffer + 1
        WordStart$ + "$"
        AddRTFTag("\cf14 ")
        SeparatorChar = 1
        Flagged       = 1
        IsString      = 1
      ElseIf IsDecNumber(WordStart$)
        AddRTFTag("\cf10 ")
        Flagged       = 1
        ;If SeparatorChar = '*' Or SeparatorChar = '%'
          UseSpecialOperator = 1
        ;EndIf
      ElseIf IsBasicKeyword(WordStart$)=0 And (SeparatorChar = '*' Or SeparatorChar = '%')
        UseSpecialOperator = 1
      ElseIf SeparatorChar = ':' And LineStart  ; Labels
        AddRTFTag("\cf13 ")
        Flagged = 1
      EndIf
    Else
      If WasSpecialOperator ;And (SeparatorChar = '*' Or SeparatorChar = '%')
        ;If (OldSeparatorChar = '@' Or OldSeparatorChar = '?' Or OldSeparatorChar = '*' Or OldSeparatorChar = '%' Or OldSeparatorChar = '$')
          UseSpecialOperator = 1
        ;EndIf
      EndIf
    EndIf

    WasSpecialOperator  = 0  ;  pointer           MOD       constant
                             ;  multiplication    binary

    If AfterSpaces Or SeparatorChar = 11 Or SeparatorChar = 1 Or SeparatorChar = ':' Or SeparatorChar = ';' Or SeparatorChar = '"' Or SeparatorChar = '$' Or SeparatorChar = '%' Or SeparatorChar = '#' Or SeparatorChar = '.' Or SeparatorChar = '('
    
      If EnableASM And LineStart And IsASMKeyword(WordStart$)
        PokeS(WordStart$, ASMKeyword$)
        AddRTFTag("\cf7 ")
        Flagged = 1
      
      ElseIf IsBasicKeyword(WordStart$)
        If SeparatorChar <> '.' Or BasicKeyword$ = "Procedure"  Or BasicKeyword$ = "ProcedureDLL" Or BasicKeyword$ = "Data" Or BasicKeyword$ = "DefType" Or BasicKeyword$ = "Declare"  Or BasicKeyword$ = "DeclareDLL" Or BasicKeyword$ = "DeclareCDLL"
          WordStart$ = BasicKeyword$
          If IsBasicKeyword_ProgramFlow(WordStart$)
            AddRTFTag("\cf15\b ")
          Else
            AddRTFTag("\cf2\b ")
          EndIf
          Flagged = 1
        EndIf
      EndIf
      
    EndIf
    
    CopyMemoryString(WordStart$)

    If Flagged
      If EditorFontStyle & #PB_FONT_BOLD
        AddRTFTag("\cf1\b ")
        LineStart = 0
      Else
        AddRTFTag("\cf1\b0 ")
        LineStart = 0
      EndIf
    ElseIf WordStart$ ; for '!' Operator:  !=bit-XOR and !=ASM
      LineStart = 0
    EndIf

    If SeparatorChar = 1
    
    ; -------------------- Strings Constants ----------------------------

    ElseIf SeparatorChar = '"'
    
      ; Add all needed missing spaces...
      
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

      AddRTFTag("\cf5 "+Chr(34))

      *TextBuffer+1 ; Skip the '"' as it could be transformed in '0'
      While *TextBuffer\b <> '"' And *TextBuffer\b <> 0 And *TextBuffer\b <> 10
     
        If *TextBuffer\b = '\' Or *TextBuffer\b = '}' Or *TextBuffer\b = '{'
          *RTFBuffer\b = '\' : *RTFBuffer+1
        EndIf

        *RTFBuffer\b = *TextBuffer\b : *RTFBuffer+1
        *TextBuffer+1
      Wend

      If *TextBuffer\b = '"'
        *RTFBuffer\b = '"' : *RTFBuffer+1
        *TextBuffer+1
      EndIf
      
      AddRTFTag("\cf1 ")
      
    ; ------------------------- Comments --------------------------------

    ElseIf SeparatorChar = ';'

      ; Add all needed missing spaces...
      ;
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

      AddRTFTag("\cf3 ;")

      *TextBuffer+1 ; Skip the ';' as it could be transformed in '0'
      While *TextBuffer\b <> 0 And *TextBuffer\b <> 10

        If *TextBuffer\b = '\' Or *TextBuffer\b = '}' Or *TextBuffer\b = '{'
          *RTFBuffer\b = '\' : *RTFBuffer+1;
        EndIf

        *RTFBuffer\b = *TextBuffer\b : *RTFBuffer+1
        *TextBuffer+1;
      Wend

      AddRTFTag("\cf1 ")

    ; -------------------- Numeric Constants ----------------------------

    ElseIf SeparatorChar = 39 ; ' = numeric constant
    
      ; Add all needed missing spaces...
      
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

      AddRTFTag("\cf4 '")

      *TextBuffer+1 ; Skip the ' as it could be transformed in '0'
      While *TextBuffer\b <> 39 And *TextBuffer\b <> 0 And *TextBuffer\b <> 10
     
        If *TextBuffer\b = '\' Or *TextBuffer\b = '}' Or *TextBuffer\b = '{'
          *RTFBuffer\b = '\' : *RTFBuffer+1
        EndIf

        *RTFBuffer\b = *TextBuffer\b : *RTFBuffer+1
        *TextBuffer+1
      Wend

      If *TextBuffer\b = 39
        *RTFBuffer\b = 39 : *RTFBuffer+1
        *TextBuffer+1
      EndIf

      WasSpecialOperator = 1 ; for Multiply ->  'A' * 2
                             ; and MOD      ->  'B' % 2

      AddRTFTag("\cf1 ")
      
    ; ------------------------- Contants --------------------------------

    ElseIf SeparatorChar = '#'
    
      ; Add all needed missing spaces...
      ;
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

      AddRTFTag("\cf4 #")

      *TextBuffer+1  ; Skip the '#' as it could be transformed in '0'
      While ValidCharacters(*TextBuffer\b & $FF)
        *RTFBuffer\b = *TextBuffer\b : *RTFBuffer+1
        *TextBuffer+1
      Wend
      If (*TextBuffer\b & $FF) = '$'
        *RTFBuffer\b = *TextBuffer\b : *RTFBuffer+1
        *TextBuffer+1
      EndIf

      WasSpecialOperator = 1 ; for Multiply ->  #const * 2
                             ; and MOD      ->  #const % 2

      AddRTFTag("\cf1 ")
    
    ; ------------------------- ASM --------------------------------

    ElseIf LineStart And SeparatorChar = '!'
    
      ; Add all needed missing spaces...
      ;
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

      AddRTFTag("\cf7 !")

      *TextBuffer+1 ; Skip the '!' as it could be transformed in '0'
      While *TextBuffer\b <> ' ' And *TextBuffer\b <> ':' And *TextBuffer\b <> ';' And *TextBuffer\b <> 0 And *TextBuffer\b <> 10
     
        If *TextBuffer\b = '\' Or *TextBuffer\b = '}' Or *TextBuffer\b = '{'
          *RTFBuffer\b = '\' : *RTFBuffer+1
        EndIf

        *RTFBuffer\b = *TextBuffer\b : *RTFBuffer+1
        *TextBuffer+1
      Wend
      
      AddRTFTag("\cf1 ")

    ; ------------------------- HEX Numbers -------------------------------

    ElseIf SeparatorChar = '$' And IsString = 0
    
      ; Add all needed missing spaces...
      ;
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

      HexFound   = 0
      NoHexFound = 0
      *ScanBuffer.BYTE = (*TextBuffer+1)
      While ValidCharacters(*ScanBuffer\b & $FF) And NoHexFound=0
        If (*ScanBuffer\b & $FF) > 47 And (*ScanBuffer\b & $FF) < 58
          HexFound = 1  ; 0-9
        ElseIf (*ScanBuffer\b & $FF) > 64 And (*ScanBuffer\b & $FF) < 71
          HexFound = 1  ; A-F
        ElseIf (*ScanBuffer\b & $FF) > 96 And (*ScanBuffer\b & $FF) < 103
          HexFound = 1  ; a-f
        Else
          NoHexFound = 1
          HexFound   = 0
        EndIf
        *ScanBuffer + 1
      Wend

      If HexFound
        AddRTFTag("\cf10 $")
        *TextBuffer+1  ; Skip the '$'
        While ValidCharacters(*TextBuffer\b & $FF)
          *RTFBuffer\b = *TextBuffer\b
          *RTFBuffer+1
          *TextBuffer+1
          WasSpecialOperator = 1
        Wend
        AddRTFTag("\cf1 ")
      Else
        ;AddRTFTag("$")
        *RTFBuffer\b = SeparatorChar
        *RTFBuffer+1
        *TextBuffer+1
      EndIf


    ; ------------------------- Label-Pointer -----------------------------

    ElseIf SeparatorChar = '?'
    
      ; Add all needed missing spaces...
      ;
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

      AddRTFTag("\cf11 ?")

      *TextBuffer+1  ; Skip the '?'
      While ValidCharacters(*TextBuffer\b & $FF)
        *RTFBuffer\b = *TextBuffer\b : *RTFBuffer+1
        *TextBuffer+1
      Wend
      WasSpecialOperator = 1
      PB_TYPE_DETECTED   = #FALSE
      AddRTFTag("\cf1 ")


    ; ------------------------- @Pointer ----------------------------------

    ElseIf SeparatorChar = '@'
    
      ; Add all needed missing spaces...
      ;
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

      *ScanBuffer.BYTE = (*TextBuffer+1)
      If (*ScanBuffer\b & $FF) = '*'
        *ScanBuffer + 1
      EndIf
      While ValidCharacters(*ScanBuffer\b & $FF) Or (*ScanBuffer\b & $FF) = ' '
        *ScanBuffer + 1
      Wend
      If (*ScanBuffer\b & $FF) = '.' Or (*ScanBuffer\b & $FF) = '$' Or (*ScanBuffer\b & $FF) = '\' Or (*ScanBuffer\b & $FF) = '('
        AddRTFTag("\cf11 @")
        *TextBuffer+1
        If (*TextBuffer\b & $FF) = '*' ; remove pointer
          *RTFBuffer\b = *TextBuffer\b
          *RTFBuffer+1 : *TextBuffer+1
          While ValidCharacters(*TextBuffer\b & $FF)
            *RTFBuffer\b = *TextBuffer\b
            *RTFBuffer+1 : *TextBuffer+1
          Wend
        EndIf
      Else
        AddRTFTag("\cf11 @")
        *TextBuffer+1  ; Skip the '@'
        If (*TextBuffer\b & $FF) = '*' ; remove pointer
          *RTFBuffer\b = *TextBuffer\b
          *RTFBuffer+1 : *TextBuffer+1
        EndIf
        While ValidCharacters(*TextBuffer\b & $FF)
          *RTFBuffer\b = *TextBuffer\b
          *RTFBuffer+1 : *TextBuffer+1
        Wend
      EndIf
      WasSpecialOperator = 1
      PB_TYPE_DETECTED   = #FALSE
      AddRTFTag("\cf1 ")

    ; ------------------------- *Pointer ---------------------------------

    ElseIf SeparatorChar = '*'
    
      ; Add all needed missing spaces...
      ;
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

      If UseSpecialOperator
        AddRTFTag("\cf8 *")
        *TextBuffer+1  ; Skip the '*'
      Else
        AddRTFTag("\cf11 *")
        *TextBuffer+1  ; Skip the '*'
        While ValidCharacters(*TextBuffer\b & $FF)
          *RTFBuffer\b = *TextBuffer\b
          *RTFBuffer+1 : *TextBuffer+1
        Wend
        WasSpecialOperator = 1
        PB_TYPE_DETECTED   = #FALSE
      EndIf

      AddRTFTag("\cf1 ")


    ; ------------------------- %binary or MOD ---------------------------

    ElseIf SeparatorChar = '%'
    
      ; Add all needed missing spaces...
      ;
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

      If UseSpecialOperator
        AddRTFTag("\cf8 %")
        *TextBuffer+1  ; Skip the '%'
        AddRTFTag("\cf1 ")
      Else
        AddRTFTag("\cf10 %")
        *TextBuffer+1  ; Skip the '%'
        
        ; highlight 0 and 1 for binary
        While (*TextBuffer\b & $FF) = '0' Or (*TextBuffer\b & $FF) = '1'
          *RTFBuffer\b = *TextBuffer\b
          *RTFBuffer+1 : *TextBuffer+1
        Wend
        AddRTFTag("\cf1 ")
        WasSpecialOperator = 1

        ; if the user types other valid chars than '0' and '1',
        ; highlight it in default color, so he can see its wrong :)
        While ValidCharacters(*TextBuffer\b & $FF)
          *RTFBuffer\b = *TextBuffer\b
          *RTFBuffer+1 : *TextBuffer+1
        Wend
      EndIf



    ; ------------------------- Operators --------------------------------

    ; = + - / & | ! ~ < >
    ElseIf SeparatorChar = '=' Or SeparatorChar = '+' Or SeparatorChar = '-' Or SeparatorChar = '/' Or SeparatorChar = '&' Or SeparatorChar = '|' Or SeparatorChar = '!' Or SeparatorChar = '~' Or SeparatorChar = '<' Or SeparatorChar = '>'

      ; Add all needed missing spaces...
      ;
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

      AddRTFTag("\cf8 ")

      *RTFBuffer\b = SeparatorChar
      *RTFBuffer+1 : *TextBuffer+1

      AddRTFTag("\cf1 ")


    ; ------------------------- Separators -------------------------------

    ; ( ) [ ] . , : \
    ElseIf SeparatorChar = '(' Or SeparatorChar = ')' Or SeparatorChar = '[' Or SeparatorChar = ']' Or SeparatorChar = '.' Or SeparatorChar = ',' Or SeparatorChar = ':' Or SeparatorChar = '\'

      ; Add all needed missing spaces...
      ;
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

       AddRTFTag("\cf12 ")

      *RTFBuffer\b = SeparatorChar
      *RTFBuffer+1 : *TextBuffer+1

      If SeparatorChar = '\'
        *RTFBuffer\b = SeparatorChar : *RTFBuffer+1
      ElseIf SeparatorChar = ')' Or SeparatorChar = ']'
        WasSpecialOperator = 1
      EndIf

      AddRTFTag("\cf1 ")


    ; --------------------- Default processing ---------------------------

    ElseIf ValidCharacters(*TextBuffer\b & $FF) = 0

      ; Add all needed missing spaces...
      ;
      While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

      ; - Detecting an escape character.., add an extra '\' To annulate it.

      If SeparatorChar = '\' Or SeparatorChar = '{' Or SeparatorChar = '}'
        *RTFBuffer\b = '\' : *RTFBuffer+1;
      EndIf
      
      *RTFBuffer\b = SeparatorChar
      *RTFBuffer+1 : *TextBuffer+1;
    EndIf
 
    ; Add all needed missing spaces...
    ;
    While (AfterSpaces) : *RTFBuffer\b = ' ' : *RTFBuffer+1 : AfterSpaces-1 : Wend

    If NewLine Or SeparatorChar = ':'
      LineStart = 1
    Else
      LineStart = 0
    EndIf
    
    
    If NewLine
      *TextBuffer+1
      *LineStart = *TextBuffer
      AddRTFTag("\line ")
      NewLine = 0
    EndIf
    
    If *RTFBuffer - *RTFBufferStart > RTFBufferSize-2000  ; If the RTF buffer is too small, just increase it..
      RTFBufferPosition = *RTFBuffer-*RTFBufferStart;
      RTFBufferSize     = RTFBufferSize+50000;
      *RTFBufferStart   = GlobalReAlloc_(*RTFBufferStart, RTFBufferSize, #GMEM_MOVEABLE);
      *RTFBuffer        = *RTFBufferStart + RTFBufferPosition;
    EndIf

    OldSeparatorChar = SeparatorChar

  Wend

  AddRTFTag("}}")

  ;If CreateFile(100, "c:\pbidetest2.rtf")
  ;  WriteData(*RTFBufferStart, *RTFBuffer - *RTFBufferStart)
  ;  MessageRequester("Info", "TestFile wrote on c:\pbidetest2.rtf")
  ;  CloseFile(100)
  ;EndIf

  ProcedureReturn *RTFBufferStart
EndProcedure


DataSection

  IID_ITextDocument:  ; {8CC497C0-A1DF-11ce-8098-00AA0047BE5D}
    Data.l $8CC497C0
    Data.w $A1DF, $11ce
    Data.b $80, $98, $00, $AA, $00, $47, $BE, $5D


  ;- Keywords - BASIC
  BasicKeywords:
    Data$ "And", "and"

    Data$ "Break", "break"

    Data$ "CallDebugger"     , "calldebugger"
    Data$ "Case"             , "case"
    Data$ "CompilerCase"     , "compilercase"
    Data$ "CompilerDefault"  , "compilerdefault"
    Data$ "CompilerElse"     , "compilerelse"
    Data$ "CompilerEndIf"    , "compilerendif"
    Data$ "CompilerEndSelect", "compilerendselect"
    Data$ "CompilerIf"       , "compilerif"
    Data$ "CompilerSelect"   , "compilerselect"
    Data$ "Continue"         , "continue"

    Data$ "Data"           , "data"
    Data$ "DataSection"    , "datasection"
    Data$ "Debug"          , "debug"
    Data$ "DebugLevel"     , "debuglevel"
    Data$ "Declare"        , "declare"
    Data$ "DeclareCDLL"    , "declarecdll"
    Data$ "DeclareDLL"     , "declaredll"
    Data$ "Default"        , "default"
    Data$ "DefType"        , "deftype"
    Data$ "Dim"            , "dim"
    Data$ "DisableDebugger", "disabledebugger"

    Data$ "Else"              , "else"
    Data$ "ElseIf"            , "elseif"
    Data$ "EnableDebugger"    , "enabledebugger"
    Data$ "End"               , "end"
    Data$ "EndDataSection"    , "enddatasection"
    Data$ "EndEnumeration"    , "endenumeration"
    Data$ "EndIf"             , "endif"
    Data$ "EndInterface"      , "endinterface"
    Data$ "EndProcedure"      , "endprocedure"
    Data$ "EndSelect"         , "endselect"
    Data$ "EndStructure"      , "endstructure"
    Data$ "EndStructureUnion" , "endstructureunion"
    Data$ "Enumeration"       , "enumeration"
    Data$ "Extends"           , "extends"

    Data$ "FakeReturn"   , "fakereturn"
    Data$ "For"          , "for"
    Data$ "ForEach"      , "foreach"
    Data$ "ForEver"      , "forever"

    Data$ "Global", "global"
    Data$ "Gosub" , "gosub"
    Data$ "Goto"  , "goto"

    Data$ "If"            , "if"
    Data$ "IncludeBinary" , "includebinary"
    Data$ "IncludeFile"   , "includefile"
    Data$ "IncludePath"   , "includepath"
    Data$ "Interface"     , "interface"


    Data$ "NewList", "newlist"
    Data$ "Next", "next"

    Data$ "Or", "or"

    Data$ "Procedure"      , "procedure"
    Data$ "ProcedureCDLL"  , "procedurecdll"
    Data$ "ProcedureDLL"   , "proceduredll"
    Data$ "ProcedureReturn", "procedurereturn"
    Data$ "Protected"      , "protected"

    Data$ "Read", "read"
    Data$ "Repeat", "repeat"
    Data$ "Restore", "restore"
    Data$ "Return", "return"

    Data$ "Select", "select"
    Data$ "Shared", "shared"
    Data$ "Static", "static"
    Data$ "Step", "step"
    Data$ "Structure", "structure"
    Data$ "StructureUnion", "structureunion"
    
    Data$ "To", "to"

    Data$ "Until", "until"

    Data$ "Wend" , "wend"
    Data$ "While", "while"

    Data$ "XIncludeFile" , "xincludefile"


  ;- Keywords - ASM
  ASMKeywords:
    Data$ "AAA"
    Data$ "AAD"
    Data$ "AAM"
    Data$ "AAS"
    Data$ "ADC"
    Data$ "ADD"
    Data$ "AND"
    Data$ "ARPL"
    
    Data$ "BOUND"
    Data$ "BSF"
    Data$ "BSR"
    Data$ "BSWAP"
    Data$ "BT"
    Data$ "BTC"
    Data$ "BTR"
    Data$ "BTS"
    
    Data$ "CALL"
    Data$ "CBW"
    Data$ "CDQ"
    Data$ "CLC"
    Data$ "CLD"
    Data$ "CLI"
    Data$ "CLTS"
    Data$ "CMC"
    Data$ "CMP"
    Data$ "CMPS"
    Data$ "CMPXCHG"
    Data$ "CWD"
    Data$ "CWDE"
    
    Data$ "DAA"
    Data$ "DAS"
    Data$ "DB"
    Data$ "DD"
    Data$ "DEC"
    Data$ "DIV"
    Data$ "DW"
    
    Data$ "ENTER"
    Data$ "ESC"
    
    Data$ "F2XM1"
    Data$ "FABS"
    Data$ "FADD"
    Data$ "FCHS"
    Data$ "FCLEX"
    Data$ "FCOM"
    Data$ "FDIV"
    Data$ "FDIVR"
    Data$ "FFREE"
    Data$ "FINCSTP"
    Data$ "FINIT"
    Data$ "FLD"
    Data$ "FLD1"
    Data$ "FLDCW"
    Data$ "FMUL"
    Data$ "FNOP"
    Data$ "FPATAN"
    Data$ "FPREM"
    Data$ "FRNDINT"
    Data$ "FSAVE"
    Data$ "FSCALE"
    Data$ "FSETPM"
    Data$ "FSIN"
    Data$ "FSQRT"
    Data$ "FST"
    Data$ "FSTENV"
    Data$ "FSTSW"
    Data$ "FSUB"
    Data$ "FSUBR"
    Data$ "FTST"
    Data$ "FUCOM"
    Data$ "FWAIT"
    Data$ "FXAM"
    Data$ "FXCH"
    Data$ "FXTRACT"
    Data$ "FYL2X"
    Data$ "FYL2XP1"
    
    Data$ "HLT"
    
    Data$ "IDIV"
    Data$ "IMUL"
    Data$ "IN"
    Data$ "INC"
    Data$ "INS"
    Data$ "INT"
    Data$ "INTO"
    Data$ "INVLPG"
    Data$ "IRET"
    Data$ "IRETD"
    
    Data$ "JA"
    Data$ "JAE"
    Data$ "JB"
    Data$ "JBE"
    Data$ "JC"
    Data$ "JCXZ"
    Data$ "JE"
    Data$ "JECXZ"
    Data$ "JG"
    Data$ "JGE"
    Data$ "JL"
    Data$ "JLE"
    Data$ "JMP"
    Data$ "JNA"
    Data$ "JNAE"
    Data$ "JNB"
    Data$ "JNBE"
    Data$ "JNC"
    Data$ "JNE"
    Data$ "JNG"
    Data$ "JNGE"
    Data$ "JNL"
    Data$ "JNLE"
    Data$ "JNO"
    Data$ "JNP"
    Data$ "JNS"
    Data$ "JNZ"
    Data$ "JO"
    Data$ "JP"
    Data$ "JPE"
    Data$ "JPO"
    Data$ "JS"
    Data$ "JZ"
    
    Data$ "LAHF"
    Data$ "LAR"
    Data$ "LDS"
    Data$ "LEA"
    Data$ "LEAVE"
    Data$ "LES"
    Data$ "LFS"
    Data$ "LGDT"
    Data$ "LGS"
    Data$ "LIDT"
    Data$ "LLDT"
    Data$ "LMSW"
    Data$ "LOCK"
    Data$ "LODS"
    Data$ "LODSB"
    Data$ "LODSW"
    Data$ "LODSD"
    Data$ "LOOP"
    Data$ "LOOPE"
    Data$ "LOOPNE"
    Data$ "LOOPNZ"
    Data$ "LOOPZ"
    Data$ "LSL"
    Data$ "LSS"
    Data$ "LTR"
    
    Data$ "MOV"
    Data$ "MOVS"
    Data$ "MOVSX"
    Data$ "MOVZX"
    Data$ "MUL"
    
    Data$ "NEG"
    Data$ "NOP"
    Data$ "NOT"
    
    Data$ "OR"
    Data$ "OUT"
    Data$ "OUTS"
    
    Data$ "POP"
    Data$ "POPA"
    Data$ "POPAD"
    Data$ "POPF"
    Data$ "POPFD"
    Data$ "PUSH"
    Data$ "PUSHA"
    Data$ "PUSHAD"
    Data$ "PUSHF"
    Data$ "PUSHFD"
    
    Data$ "RCL"
    Data$ "RCR"
    Data$ "REP"
    Data$ "REPE"
    Data$ "REPNE"
    Data$ "REPNZ"
    Data$ "REPZ"
    Data$ "RET"
    Data$ "RETF"
    Data$ "ROL"
    Data$ "ROR"
    
    Data$ "SAHF"
    Data$ "SAL"
    Data$ "SAR"
    Data$ "SBB"
    Data$ "SCAS"
    Data$ "SETAE"
    Data$ "SETB"
    Data$ "SETBE"
    Data$ "SETC"
    Data$ "SETE"
    Data$ "SETG"
    Data$ "SETGE"
    Data$ "SETL"
    Data$ "SETLE"
    Data$ "SETNA"
    Data$ "SETNAE"
    Data$ "SETNB"
    Data$ "SETNC"
    Data$ "SETNE"
    Data$ "SETNG"
    Data$ "SETNGE"
    Data$ "SETNL"
    Data$ "SETNLE"
    Data$ "SETNO"
    Data$ "SETNP"
    Data$ "SETNS"
    Data$ "SETNZ"
    Data$ "SETO"
    Data$ "SETP"
    Data$ "SETPE"
    Data$ "SETPO"
    Data$ "SETS"
    Data$ "SETZ"
    Data$ "SGDT"
    Data$ "SHL"
    Data$ "SHLD"
    Data$ "SHR"
    Data$ "SHRD"
    Data$ "SIDT"
    Data$ "SLDT"
    Data$ "SMSW"
    Data$ "STC"
    Data$ "STD"
    Data$ "STI"
    Data$ "STOS"
    Data$ "STR"
    Data$ "SUB"
    
    Data$ "TEST"
    
    Data$ "VERR"
    Data$ "VERW"
    
    Data$ "WAIT"
    Data$ "WBINVD"
    
    Data$ "XCHG"
    Data$ "XLAT"
    Data$ "XLATB"
    Data$ "XOR"

EndDataSection
; ExecutableFormat=
; EOF